from __future__ import division

import numpy as np
from scipy import special
from scipy.special import erf

from ..utils import utils
from ..core.constants import CONSTANTS
from ..core.modeling_framework import ModelProperties
from ..core.signal_model_properties import AnisotropicSignalModelProperties
from dipy.utils.optpkg import optional_package

numba, have_numba, _ = optional_package("numba")

DIFFUSIVITY_SCALING = 1e-9
DIAMETER_SCALING = 1e-6
A_SCALING = 1e-12


__all__ = [
    'C1Stick',
    'C2CylinderStejskalTannerApproximation',
    'C3CylinderCallaghanApproximation',
    'C4CylinderGaussianPhaseApproximation'
]


class C1Stick(ModelProperties, AnisotropicSignalModelProperties):
    r""" The Stick model [1]_ - a cylinder with zero radius - typically used
    for intra-axonal diffusion.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    lambda_par : float,
        parallel diffusivity in m^2/s.

    References
    ----------
    .. [1] Behrens et al.
           "Characterization and propagation of uncertainty in
            diffusion-weighted MR imaging"
           Magnetic Resonance in Medicine (2003)
    """

    _required_acquisition_parameters = ['bvalues', 'gradient_directions']

    _parameter_ranges = {
        'mu': ([0, np.pi], [-np.pi, np.pi]),
        'lambda_par': (.1, 3)
    }
    _parameter_scales = {
        'mu': np.r_[1., 1.],
        'lambda_par': DIFFUSIVITY_SCALING
    }
    _parameter_types = {
        'mu': 'orientation',
        'lambda_par': 'normal'
    }
    _model_type = 'CompartmentModel'

    def __init__(self, mu=None, lambda_par=None):
        self.mu = mu
        self.lambda_par = lambda_par

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Estimates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        bvals = acquisition_scheme.bvalues
        n = acquisition_scheme.gradient_directions

        lambda_par_ = kwargs.get('lambda_par', self.lambda_par)
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)
        E_stick = _attenuation_parallel_stick(bvals, lambda_par_, n, mu)
        return E_stick

    def spherical_mean(self, acquisition_scheme, **kwargs):
        """
        Estimates spherical mean for every shell in acquisition scheme for
        Stick model.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E_mean : array of size (Nshells)
            spherical mean of the Stick model for every acquisition shell.
        """
        bvals = acquisition_scheme.shell_bvalues
        bvals_ = bvals[~acquisition_scheme.shell_b0_mask]

        lambda_par = kwargs.get('lambda_par', self.lambda_par)

        E_mean = np.ones_like(bvals)
        bval_indices_above0 = bvals > 0
        bvals_ = bvals[bval_indices_above0]
        E_mean_ = ((np.sqrt(np.pi) * erf(np.sqrt(bvals_ * lambda_par))) /
                   (2 * np.sqrt(bvals_ * lambda_par)))
        E_mean[bval_indices_above0] = E_mean_
        return E_mean


class C2CylinderStejskalTannerApproximation(
        ModelProperties, AnisotropicSignalModelProperties):
    r""" The Stejskal-Tanner approximation of the cylinder model with finite
    radius, proposed by Soderman and Jonsson [1]_. Assumes that both the short
    gradient pulse (SGP) approximation is met and long diffusion time limit is
    reached. The perpendicular cylinder diffusion therefore only depends on the
    q-value of the acquisition.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    lambda_par : float,
        parallel diffusivity in m^2/s.
    diameter : float,
        cylinder diameter in meters.

    Returns
    -------
    E : array, shape (N,)
        signal attenuation

    References
    ----------
    .. [1] Soderman, Olle, and Bengt Jonsson. "Restricted diffusion in
            cylindrical geometry." Journal of Magnetic Resonance, Series A
            117.1 (1995): 94-97.
    """

    _required_acquisition_parameters = [
        'bvalues', 'gradient_directions', 'qvalues']

    _parameter_ranges = {
        'mu': ([0, np.pi], [-np.pi, np.pi]),
        'lambda_par': (.1, 3),
        'diameter': (1e-2, 20)
    }
    _parameter_scales = {
        'mu': np.r_[1., 1.],
        'lambda_par': DIFFUSIVITY_SCALING,
        'diameter': DIAMETER_SCALING
    }
    _parameter_types = {
        'mu': 'orientation',
        'lambda_par': 'normal',
        'diameter': 'cylinder'
    }
    _model_type = 'CompartmentModel'

    def __init__(
        self,
        mu=None, lambda_par=None,
        diameter=None
    ):
        self.mu = mu
        self.lambda_par = lambda_par
        self.diameter = diameter

    def perpendicular_attenuation(
        self, q, diameter
    ):
        "Returns the cylinder's perpendicular signal attenuation."
        radius = diameter / 2
        # Eq. [6] in the paper
        E = ((2 * special.jn(1, 2 * np.pi * q * radius)) ** 2 /
             (2 * np.pi * q * radius) ** 2)
        return E

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Estimates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        bvals = acquisition_scheme.bvalues
        n = acquisition_scheme.gradient_directions
        q = acquisition_scheme.qvalues

        diameter = kwargs.get('diameter', self.diameter)
        lambda_par = kwargs.get('lambda_par', self.lambda_par)
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)
        mu_perpendicular_plane = np.eye(3) - np.outer(mu, mu)
        magnitude_perpendicular = np.linalg.norm(
            np.dot(mu_perpendicular_plane, n.T),
            axis=0
        )
        E_parallel = _attenuation_parallel_stick(bvals, lambda_par, n, mu)
        E_perpendicular = np.ones_like(q)
        q_perp = q * magnitude_perpendicular
        q_nonzero = q_perp > 0  # only q>0 attenuate
        E_perpendicular[q_nonzero] = self.perpendicular_attenuation(
            q_perp[q_nonzero], diameter
        )
        return E_parallel * E_perpendicular


class C3CylinderCallaghanApproximation(
        ModelProperties, AnisotropicSignalModelProperties):
    r""" The Callaghan model [1]_ - a cylinder with finite radius - typically
    used for intra-axonal diffusion. The perpendicular diffusion is modelled
    after Callaghan's solution for the disk. Is dependent on both q-value
    and diffusion time.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    lambda_par : float,
        parallel diffusivity in m^2/s.
    diameter : float,
        cylinder (axon) diameter in meters.
    diffusion_perpendicular : float,
        the intra-cylindrical, perpenicular diffusivity. By default it is set
        to a typical value for intra-axonal diffusion as 1.7e-9 m^2/s.
    number_of_roots : integer,
        number of roots to use for the Callaghan cylinder model.
    number_of_function : integer,
        number of functions to use for the Callaghan cylinder model.

    References
    ----------
    .. [1] Callaghan, Paul T. "Pulsed-gradient spin-echo NMR for planar,
            cylindrical, and spherical pores under conditions of wall
            relaxation." Journal of magnetic resonance, Series A 113.1 (1995):
            53-59.
    """

    _required_acquisition_parameters = [
        'bvalues', 'gradient_directions', 'qvalues', 'tau']

    _parameter_ranges = {
        'mu': ([0, np.pi], [-np.pi, np.pi]),
        'lambda_par': (.1, 3),
        'diameter': (1e-2, 20)
    }
    _parameter_scales = {
        'mu': np.r_[1., 1.],
        'lambda_par': DIFFUSIVITY_SCALING,
        'diameter': DIAMETER_SCALING
    }
    _parameter_types = {
        'mu': 'orientation',
        'lambda_par': 'normal',
        'diameter': 'cylinder'
    }
    _model_type = 'CompartmentModel'

    def __init__(
        self,
        mu=None, lambda_par=None,
        diameter=None,
        diffusion_perpendicular=CONSTANTS['water_in_axons_diffusion_constant'],
        number_of_roots=20,
        number_of_functions=50,
    ):
        self.mu = mu
        self.lambda_par = lambda_par
        self.diffusion_perpendicular = diffusion_perpendicular
        self.diameter = diameter

        self.alpha = np.empty((number_of_roots, number_of_functions))
        self.alpha[0, 0] = 0
        if number_of_roots > 1:
            self.alpha[1:, 0] = special.jnp_zeros(0, number_of_roots - 1)
        for m in range(1, number_of_functions):
            self.alpha[:, m] = special.jnp_zeros(m, number_of_roots)

    def perpendicular_attenuation(self, q, tau, diameter):
        "Implements the finite time Callaghan model for cylinders"
        radius = diameter / 2.
        alpha = self.alpha
        q_argument = 2 * np.pi * q * radius
        q_argument_2 = q_argument ** 2
        res = np.zeros_like(q)

        J = special.j1(q_argument) ** 2
        for k in range(0, self.alpha.shape[0]):
            alpha2 = alpha[k, 0] ** 2
            update = (
                4 * np.exp(-alpha2 * self.diffusion_perpendicular *
                           tau / radius ** 2) *
                q_argument_2 /
                (q_argument_2 - alpha2) ** 2 * J
            )
            res += update

        for m in range(1, self.alpha.shape[1]):
            J = special.jvp(m, q_argument, 1)
            q_argument_J = (q_argument * J) ** 2
            for k in range(self.alpha.shape[0]):
                alpha2 = self.alpha[k, m] ** 2
                update = (
                    8 * np.exp(-alpha2 * self.diffusion_perpendicular *
                               tau / radius ** 2) *
                    alpha2 / (alpha2 - m ** 2) *
                    q_argument_J /
                    (q_argument_2 - alpha2) ** 2
                )
                res += update
        return res

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Estimates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        bvals = acquisition_scheme.bvalues
        n = acquisition_scheme.gradient_directions
        q = acquisition_scheme.qvalues
        tau = acquisition_scheme.tau

        diameter = kwargs.get('diameter', self.diameter)
        lambda_par = kwargs.get('lambda_par', self.lambda_par)
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)
        mu_perpendicular_plane = np.eye(3) - np.outer(mu, mu)
        magnitude_perpendicular = np.linalg.norm(
            np.dot(mu_perpendicular_plane, n.T),
            axis=0
        )
        E_parallel = _attenuation_parallel_stick(bvals, lambda_par, n, mu)
        E_perpendicular = np.ones_like(q)
        q_perp = q * magnitude_perpendicular

        q_nonzero = q_perp > 0
        E_perpendicular[q_nonzero] = self.perpendicular_attenuation(
            q_perp[q_nonzero], tau[q_nonzero], diameter
        )
        return E_parallel * E_perpendicular


class C4CylinderGaussianPhaseApproximation(
        ModelProperties, AnisotropicSignalModelProperties):
    r""" The Gaussian phase model [1]_ - a cylinder with finite radius -
    typically used for intra-axonal diffusion. The perpendicular diffusion is
    modelled after Van Gelderen's solution for the disk. It is dependent on
    gradient strength, pulse separation and pulse length.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    lambda_par : float,
        parallel diffusivity in 10^9 m^2/s.
    diameter : float,
        cylinder (axon) diameter in meters.


    References
    ----------
    .. [1] Van Gelderen et al. "Evaluation of Restricted Diffusion in
            Cylinders. Phosphocreatine in Rabbit Leg Muscle"
            Journal of Magnetic Resonance Series B (1994)
    """

    _required_acquisition_parameters = [
        'bvalues', 'gradient_directions',
        'gradient_strengths', 'delta', 'Delta']

    _parameter_ranges = {
        'mu': ([0, np.pi], [-np.pi, np.pi]),
        'lambda_par': (.1, 3),
        'diameter': (1e-2, 20)
    }
    _parameter_scales = {
        'mu': np.r_[1., 1.],
        'lambda_par': DIFFUSIVITY_SCALING,
        'diameter': DIAMETER_SCALING
    }
    _parameter_types = {
        'mu': 'orientation',
        'lambda_par': 'normal',
        'diameter': 'cylinder'
    }
    _model_type = 'CompartmentModel'
    _CYLINDER_TRASCENDENTAL_ROOTS = np.sort(special.jnp_zeros(1, 100))

    def __init__(
        self,
        mu=None, lambda_par=None,
        diameter=None,
        diffusion_perpendicular=CONSTANTS['water_in_axons_diffusion_constant']
    ):
        self.mu = mu
        self.lambda_par = lambda_par
        self.diffusion_perpendicular = diffusion_perpendicular
        self.gyromagnetic_ratio = CONSTANTS['water_gyromagnetic_ratio']
        self.diameter = diameter

    def perpendicular_attenuation(
        self, gradient_strength, delta, Delta, diameter
    ):
        "Calculates the cylinder's perpendicular signal attenuation."
        D = self.diffusion_perpendicular
        gamma = self.gyromagnetic_ratio
        return _attenuation_perpendicular_gaussian_phase(
            diameter, gradient_strength, delta, Delta,
            D, gamma, self._CYLINDER_TRASCENDENTAL_ROOTS)

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        bvals = acquisition_scheme.bvalues
        n = acquisition_scheme.gradient_directions
        g = acquisition_scheme.gradient_strengths
        delta = acquisition_scheme.delta
        Delta = acquisition_scheme.Delta

        diameter = kwargs.get('diameter', self.diameter)
        lambda_par = kwargs.get('lambda_par', self.lambda_par)
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)
        mu_perpendicular_plane = np.eye(3) - np.outer(mu, mu)
        magnitude_perpendicular = np.linalg.norm(
            np.dot(mu_perpendicular_plane, n.T),
            axis=0
        )
        E_parallel = _attenuation_parallel_stick(bvals, lambda_par, n, mu)
        E_perpendicular = np.ones_like(g)
        g_perp = g * magnitude_perpendicular

        g_nonzero = g_perp > 0
        # for every unique combination get the perpendicular attenuation
        unique_deltas = np.unique([acquisition_scheme.shell_delta,
                                   acquisition_scheme.shell_Delta], axis=1)
        for delta_, Delta_ in zip(*unique_deltas):
            mask = np.all([g_nonzero, delta == delta_, Delta == Delta_],
                          axis=0)
            E_perpendicular[mask] = self.perpendicular_attenuation(
                g_perp[mask], delta_, Delta_, diameter
            )
        return E_parallel * E_perpendicular


def _attenuation_parallel_stick(bvals, lambda_par, n, mu):
    "Free gaussian diffusion for parallel cylinder direction."
    return np.exp(-bvals * lambda_par * np.dot(n, mu) ** 2)


def _attenuation_perpendicular_gaussian_phase(
        diameter, gradient_strength, delta, Delta,
        D, gamma, CYLINDER_TRASCENDENTAL_ROOTS):
    "Perpendicular Gaussian Phase signal attenuation."
    radius = diameter / 2.
    first_factor = -2 * (gradient_strength * gamma) ** 2
    alpha = CYLINDER_TRASCENDENTAL_ROOTS / radius
    alpha2 = alpha ** 2
    alpha2D = alpha2 * D

    summands = (
        2 * alpha2D * delta - 2 +
        2 * np.exp(-alpha2D * delta) +
        2 * np.exp(-alpha2D * Delta) -
        np.exp(-alpha2D * (Delta - delta)) -
        np.exp(-alpha2D * (Delta + delta))
    ) / (D ** 2 * alpha ** 6 * (radius ** 2 * alpha2 - 1))

    E = np.exp(first_factor * summands.sum())
    return E


if have_numba:
    _attenuation_parallel_stick = numba.njit()(_attenuation_parallel_stick)
    _attenuation_perpendicular_gaussian_phase = numba.njit()(
        _attenuation_perpendicular_gaussian_phase)



from ..core.modeling_framework import ModelProperties
from ..core.signal_model_properties import IsotropicSignalModelProperties
from ..core.constants import CONSTANTS
from scipy import special
import numpy as np

DIAMETER_SCALING = 1e-6

__all__ = [
    'S1Dot',
    'S2SphereStejskalTannerApproximation'
]


class S1Dot(ModelProperties, IsotropicSignalModelProperties):
    r"""
    The Dot model [1]_ - an non-diffusing compartment.
    It has no parameters and returns 1 no matter the input.

    References
    ----------
    .. [1] Panagiotaki et al.
           "Compartment models of the diffusion MR signal in brain white
            matter: a taxonomy and comparison". NeuroImage (2012)
    """
    _required_acquisition_parameters = []

    _parameter_ranges = {
    }
    _parameter_scales = {
    }
    _parameter_types = {
    }
    _model_type = 'CompartmentModel'

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        E_dot = np.ones(acquisition_scheme.number_of_measurements)
        return E_dot


class S2SphereStejskalTannerApproximation(
        ModelProperties, IsotropicSignalModelProperties):
    r"""
    The Stejskal Tanner signal approximation of a sphere model. It assumes
    that pulse length is infinitessimally small and diffusion time large enough
    so that the diffusion is completely restricted. Only depends on q-value.

    Parameters
    ----------
    diameter : float,
        sphere diameter in meters.

    References
    ----------
    .. [1] Balinov, Balin, et al. "The NMR self-diffusion method applied to
        restricted diffusion. Simulation of echo attenuation from molecules in
        spheres and between planes." Journal of Magnetic Resonance, Series A
        104.1 (1993): 17-25.
    """
    _required_acquisition_parameters = ['qvalues']

    _parameter_ranges = {
        'diameter': (1e-2, 20)
    }
    _parameter_scales = {
        'diameter': DIAMETER_SCALING
    }
    _parameter_types = {
        'diameter': 'sphere',
    }
    _model_type = 'CompartmentModel'

    def __init__(self, diameter=None):
        self.diameter = diameter

    def sphere_attenuation(self, q, diameter):
        "The signal attenuation for the sphere model."
        radius = diameter / 2
        factor = 2 * np.pi * q * radius
        E = (
            3 / (factor ** 2) *
            (
                np.sin(factor) / factor -
                np.cos(factor)
            )
        ) ** 2
        return E

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        q = acquisition_scheme.qvalues
        diameter = kwargs.get('diameter', self.diameter)
        E_sphere = np.ones_like(q)
        q_nonzero = q > 0  # only q>0 attenuate
        E_sphere[q_nonzero] = self.sphere_attenuation(
            q[q_nonzero], diameter)
        return E_sphere


class _S3SphereCallaghanApproximation(
        ModelProperties, IsotropicSignalModelProperties):
    r"""
    The Callaghan model [1]_ of diffusion inside a sphere.

    Parameters
    ----------
    diameter : float
        Distance between the two plates in meters.
    diffusion_constant : float,
        The diffusion constant of the water particles between the two planes.
        The default value is the approximate diffusivity of water inside axons
        as 1.7e-9 m^2/s.
    number_of_roots : integer,
        The number of roots for the Callaghan approximation.

    References
    ----------
    [1] Callaghan, "Pulsed-Gradient Spin-Echo NMR for Planar, Cylindrical,
        and Spherical Pores under Conditions of Wall Relaxation", JMR 1995
    """
    _required_acquisition_parameters = ['qvalues', 'tau']

    _parameter_ranges = {
        'diameter': (1e-2, 20)
    }

    _parameter_scales = {
        'diameter': DIAMETER_SCALING
    }

    _parameter_types = {
        'diameter': 'sphere'
    }
    _model_type = 'CompartmentModel'

    def __init__(
        self,
        diameter=None,
        diffusion_constant=CONSTANTS['water_in_axons_diffusion_constant'],
        number_of_roots=20,
        number_of_functions=50,
    ):

        self.diameter = diameter
        self.Dintra = diffusion_constant
        self.alpha = np.empty((number_of_roots, number_of_functions))
        self.alpha[0, 0] = 0
        if number_of_roots > 1:
            self.alpha[1:, 0] = special.jnp_zeros(0, number_of_roots - 1)
        for m in range(1, number_of_functions):
            self.alpha[:, m] = special.jnp_zeros(m, number_of_roots)

    def sphere_attenuation(self, q, tau, diameter):
        """Implements the finite time Callaghan model for planes."""
        radius = diameter / 2.0
        q_argument = 2 * np.pi * q * radius
        q_argument_2 = q_argument ** 2
        res = np.zeros_like(q)

        # J = special.spherical_jn(q_argument)
        Jder = special.spherical_jn(q_argument, derivative=True)
        for k in range(0, self.alpha.shape[0]):
            for n in range(0, self.alpha.shape[1]):
                a_nk2 = self.alpha[k, n] ** 2
                update = np.exp(-a_nk2 * self.Dintra * tau / radius ** 2)
                update *= ((2 * n + 1) * a_nk2) / \
                    (a_nk2 - (n - 0.5) ** 2 + 0.25)
                update *= q_argument * Jder
                update /= (q_argument_2 - a_nk2) ** 2
                res += update
        return res

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        q = acquisition_scheme.qvalues
        tau = acquisition_scheme.tau
        diameter = kwargs.get('diameter', self.diameter)

        E_sphere = np.ones_like(q)
        q_nonzero = q > 0
        E_sphere[q_nonzero] = self.sphere_attenuation(
            q[q_nonzero], tau[q_nonzero], diameter
        )
        return E_sphere


class S4SphereGaussianPhaseApproximation(
        ModelProperties, IsotropicSignalModelProperties):
    r"""
    The gaussian phase approximation for diffusion inside a sphere according
    to [1]_. It is dependent on gradient strength, pulse separation and pulse
    length.

    References
    ----------
    .. [1] Balinov, Balin, et al. "The NMR self-diffusion method applied to
        restricted diffusion. Simulation of echo attenuation from molecules in
        spheres and between planes." Journal of Magnetic Resonance, Series A
        104.1 (1993): 17-25.
    """
    _required_acquisition_parameters = ['gradient_strengths', 'delta', 'Delta']

    _parameter_ranges = {
        'diameter': (1e-2, 20)
    }
    _parameter_scales = {
        'diameter': DIAMETER_SCALING
    }
    _parameter_types = {
        'diameter': 'sphere'
    }
    _model_type = 'CompartmentModel'

    # According to Balinov et al., solutions of
    # 1/(alpha * R) * J(3/2,alpha * R) = J(5/2, alpha * R)
    # with R = 1 with alpha * R < 100 * pi
    SPHERE_TRASCENDENTAL_ROOTS = np.r_[
        # 0.,
        2.081575978, 5.940369990, 9.205840145,
        12.40444502, 15.57923641, 18.74264558, 21.89969648,
        25.05282528, 28.20336100, 31.35209173, 34.49951492,
        37.64596032, 40.79165523, 43.93676147, 47.08139741,
        50.22565165, 53.36959180, 56.51327045, 59.65672900,
        62.80000055, 65.94311190, 69.08608495, 72.22893775,
        75.37168540, 78.51434055, 81.65691380, 84.79941440,
        87.94185005, 91.08422750, 94.22655255, 97.36883035,
        100.5110653, 103.6532613, 106.7954217, 109.9375497,
        113.0796480, 116.2217188, 119.3637645, 122.5057870,
        125.6477880, 128.7897690, 131.9317315, 135.0736768,
        138.2156061, 141.3575204, 144.4994207, 147.6413080,
        150.7831829, 153.9250463, 157.0668989, 160.2087413,
        163.3505741, 166.4923978, 169.6342129, 172.7760200,
        175.9178194, 179.0596116, 182.2013968, 185.3431756,
        188.4849481, 191.6267147, 194.7684757, 197.9102314,
        201.0519820, 204.1937277, 207.3354688, 210.4772054,
        213.6189378, 216.7606662, 219.9023907, 223.0441114,
        226.1858287, 229.3275425, 232.4692530, 235.6109603,
        238.7526647, 241.8943662, 245.0360648, 248.1777608,
        251.3194542, 254.4611451, 257.6028336, 260.7445198,
        263.8862038, 267.0278856, 270.1695654, 273.3112431,
        276.4529189, 279.5945929, 282.7362650, 285.8779354,
        289.0196041, 292.1612712, 295.3029367, 298.4446006,
        301.5862631, 304.7279241, 307.8695837, 311.0112420,
        314.1528990
    ]

    def __init__(
        self, diameter=None,
        diffusion_constant=CONSTANTS['water_in_axons_diffusion_constant'],
    ):
        self.diffusion_constant = diffusion_constant
        self.gyromagnetic_ratio = CONSTANTS['water_gyromagnetic_ratio']
        self.diameter = diameter

    def sphere_attenuation(
        self, gradient_strength, delta, Delta, diameter
    ):
        "Calculates the sphere signal attenuation."

        D = self.diffusion_constant
        gamma = self.gyromagnetic_ratio
        radius = diameter / 2

        alpha = self.SPHERE_TRASCENDENTAL_ROOTS / radius
        alpha2 = alpha ** 2
        alpha2D = alpha2 * D

        first_factor = -2 * (gamma * gradient_strength) ** 2 / D
        summands = (
            alpha ** (-4) / (alpha2 * radius ** 2 - 2) *
            (
                2 * delta - (
                    2 +
                    np.exp(-alpha2D * (Delta - delta)) -
                    2 * np.exp(-alpha2D * delta) -
                    2 * np.exp(-alpha2D * Delta) +
                    np.exp(-alpha2D * (Delta + delta))
                ) / (alpha2D)
            )
        )
        E = np.exp(
            first_factor *
            summands.sum()
        )
        return E

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        g = acquisition_scheme.gradient_strengths
        delta = acquisition_scheme.delta
        Delta = acquisition_scheme.Delta

        diameter = kwargs.get('diameter', self.diameter)
        E_sphere = np.ones_like(g)

        g_nonzero = g > 0
        # for every unique combination get the perpendicular attenuation
        unique_deltas = np.unique([acquisition_scheme.shell_delta,
                                   acquisition_scheme.shell_Delta], axis=1)
        for delta_, Delta_ in zip(*unique_deltas):
            mask = np.all([g_nonzero, delta == delta_, Delta == Delta_],
                          axis=0)
            E_sphere[mask] = self.sphere_attenuation(
                g[mask], delta_, Delta_, diameter
            )
        return E_sphere



import numpy as np
from scipy.optimize import differential_evolution, minimize, fmin_cobyla

__all__ = [
    'MixOptimizer',
    'nested_to_normalized_fractions',
    'cobyla_positivity_constraint',
    'cobyla_unity_constraint'
]


class MixOptimizer:
    """ The stochastic Microstructure In Crossings (MIX) optimizer [1] uses a
    three-step process to fit the parameters of a multi-compartment (MC) model
    to data. The key innovation is that they separate linear from non-linear
    parameters in the fitting process, meaning the linear volume fractions
    and non-linear other ones(e.g. diffusivities) are optimized at different
    stages in the process.

    In the first step [1] describes using a genetic algorithm to estimate the
    non-linear parameters of an MC model. For this we use scipy's
    differential_evolution algorithm.

    In the second step [1] describes using CVX to estimate the linear volume
    fractions of an MC model. For this we use scipy's COBYLA algorithm since
    it allows us to impose the parameter constraints we need for volume
    fractions; namely that they are positive and sum up to one.

    The third and last step in [1] is a refining step to find a local minimum
    given the solutions of step one and two. For this we use scipy's
    gradient-based L-BFGS-B algorithm with nested volume fractions.

    The final result takes a model's parameter_ranges into account, only
    yielding parameters within their allowed optimization domain.

    Parameters
    ----------
    model : MultiCompartmentModel instance,
        A multicompartment model that has been instantiated using dMipy.
    acquisition_scheme : DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using dMipy.
    maxiter : integer
        The maximum allowed iterations for the differential evolution algorithm

    References
    ----------
    .. [1] Farooq, Hamza, et al. "Microstructure Imaging of Crossing (MIX)
           White Matter Fibers from diffusion MRI." Nature Scientific
           reports 6 (2016).

    """

    def __init__(self, model, acquisition_scheme, maxiter=150):
        self.model = model
        self.acquisition_scheme = acquisition_scheme
        self.maxiter = maxiter
        self.Nmodels = len(self.model.models)

    def __call__(self, data, x0_vector=np.array([np.nan])):
        """ The fitting function of the MIX algorithm. Fits the data in three
        distinct steps, first fitting non-linear parameters using
        differential_evolution, then linear parameters using COBYLA, and
        finally refining using L-BFGS-B. See main documentation for details.

        Parameters
        ----------
        data : Array of size (Ndata)
            The normalized dMRI signal attenuation to be fitted.
        x0_vector : Array of size (Nparameters)
            Possible initial guess vector for parameter initiation.

        Returns
        -------
        fitted_parameters : Array of size (Nparameters),
            The fitted MC model parameters using MIX.

        """
        # if there is only one model then MIX only uses DE.
        bounds = list(self.model.bounds_for_optimization)
        if self.Nmodels == 1:
            bounds_de = bounds
            opt_flag_array = self.model.opt_params_for_optimization
            if x0_vector is not None:
                if not np.all(np.isnan(x0_vector)):
                    bounds_de = []
                    for i, x0_ in enumerate(x0_vector):
                        if np.isnan(x0_):
                            bounds_de.append(bounds[i])

            # step 1: stochastic optimization on non-linear parameters.
            res_de = differential_evolution(
                self.stochastic_objective_function,
                bounds=bounds_de,
                maxiter=self.maxiter,
                args=(data, self.acquisition_scheme, x0_vector),
                polish=True).x
            if np.all(np.isnan(x0_vector)):
                fitted_parameters = res_de
                return fitted_parameters
            else:
                x0_bool_array = ~np.isnan(x0_vector)
                fitted_parameters = np.ones(len(x0_bool_array))
                fitted_parameters[~x0_bool_array] = res_de
                fitted_parameters[x0_bool_array] = x0_vector[
                    x0_bool_array]
                return fitted_parameters
        # if there is more than 1 model then we do the 3 steps.
        if self.Nmodels > 1:
            bounds_de = bounds
            bounds_minimize = bounds[:-1]  # nested volume fractions
            opt_flag_array = self.model.opt_params_for_optimization
            if x0_vector is not None:
                if not np.all(np.isnan(x0_vector)):
                    bounds_de = []
                    bounds_minimize = []
                    for i, x0_ in enumerate(x0_vector[:-self.Nmodels]):
                        if np.isnan(x0_):
                            bounds_de.append(bounds[i])
                    for i, x0_ in enumerate(x0_vector):
                        if opt_flag_array[i] is True:
                            bounds_minimize.append(bounds[i])
                    bounds_minimize = bounds_minimize[:-1]
                else:
                    bounds_de = bounds_de[:-self.Nmodels]
            # step 1: stochastic optimization on non-linear parameters.
            res_de = differential_evolution(self.stochastic_objective_function,
                                            bounds=bounds_de,
                                            maxiter=self.maxiter,
                                            args=(data,
                                                  self.acquisition_scheme,
                                                  x0_vector))
            optimized_parameter_vector = res_de.x
            if np.all(np.isnan(x0_vector)):
                parameter_vector = np.r_[optimized_parameter_vector,
                                         np.ones(self.Nmodels)]
            elif np.all(~np.isnan(x0_vector[-self.Nmodels:])):
                x0_bool_array = ~np.isnan(x0_vector)
                parameter_vector = np.ones(len(x0_bool_array))
                parameter_vector[~x0_bool_array] = optimized_parameter_vector
                parameter_vector[x0_bool_array] = x0_vector[x0_bool_array]
                return parameter_vector
            else:
                x0_bool_array = ~np.isnan(x0_vector)
                parameter_vector = np.ones(len(x0_bool_array))
                x0_bool_n0_vf = x0_bool_array[:-self.Nmodels]
                parameter_vector_no_vf = np.empty(
                    len(x0_bool_n0_vf), dtype=float)
                parameter_vector_no_vf[~x0_bool_n0_vf] = (
                    optimized_parameter_vector)
                parameter_vector_no_vf[x0_bool_n0_vf] = x0_vector[
                    x0_bool_array]
                parameter_vector[:-self.Nmodels] = parameter_vector_no_vf
            parameters = self.model.parameter_vector_to_parameters(
                parameter_vector * self.model.scales_for_optimization)

            # step 2: Estimating linear variables using COBYLA
            phi = self.model(self.acquisition_scheme,
                             quantity="stochastic cost function", **parameters)
            try:
                phi_inv = np.dot(np.linalg.inv(np.dot(phi.T, phi)), phi.T)
                vf_x0 = np.dot(phi_inv, data)
                vf_x0 /= np.sum(np.clip(vf_x0, 0, np.inf))
                vf = fmin_cobyla(self.cobyla_cost_function, x0=vf_x0,
                                 cons=[cobyla_positivity_constraint,
                                       cobyla_unity_constraint],
                                 args=(phi, data),
                                 maxfun=2000)
            except np.linalg.linalg.LinAlgError:
                # happens when models have the same signal attenuations.
                vf = np.ones(self.Nmodels) / float(self.Nmodels)
            vf_nested = np.ones(len(vf) - 1)
            vf_nested[0] = vf[0]
            for i in np.arange(1, len(vf_nested)):
                vf_nested[i] = vf[i] / vf[i - 1]

            # Convert to nested volume fractions
            x0_refine = np.r_[optimized_parameter_vector, vf_nested]
            # step 3: refine using gradient method
            x_fine_nested = minimize(self.objective_function, x0_refine,
                                     (data,
                                      self.acquisition_scheme,
                                      x0_vector),
                                     bounds=bounds_minimize).x
            nested_fractions = x_fine_nested[-(self.Nmodels - 1):]
            normalized_fractions = nested_to_normalized_fractions(
                nested_fractions)
            minimize_fitted_parameters = np.r_[
                x_fine_nested[:-(self.Nmodels - 1)], normalized_fractions]

            if not np.all(self.model.opt_params_for_optimization):
                x0_bool_array = ~np.isnan(x0_vector)
                fitted_parameters = np.ones(len(x0_bool_array))
                fitted_parameters[~x0_bool_array] = minimize_fitted_parameters
                fitted_parameters[x0_bool_array] = x0_vector[x0_bool_array]
            else:
                fitted_parameters = minimize_fitted_parameters
            return fitted_parameters

    def stochastic_objective_function(self, optimized_parameter_vector,
                                      data, acquisition_scheme, x0_params):
        """Objective function for stochastic non-linear parameter estimation
        using differential_evolution
        """
        x0_bool_array = ~np.isnan(x0_params)
        if self.Nmodels == 1:
            # add fixed parameters if given.
            if np.all(np.isnan(x0_params)):
                parameter_vector = optimized_parameter_vector
            else:
                parameter_vector = np.empty(len(x0_bool_array))
                parameter_vector[~x0_bool_array] = optimized_parameter_vector
                parameter_vector[x0_bool_array] = x0_params[x0_bool_array]

            parameter_vector = (
                parameter_vector * self.model.scales_for_optimization)
            parameters = self.model.parameter_vector_to_parameters(
                parameter_vector)
            E_hat = self.model(acquisition_scheme, **parameters)
        elif self.Nmodels > 1:
            if np.all(np.isnan(x0_params)):
                parameter_vector = np.r_[optimized_parameter_vector,
                                         np.ones(self.Nmodels)]
            else:
                parameter_vector = np.ones(len(x0_bool_array))
                x0_bool_n0_vf = x0_bool_array[:-self.Nmodels]
                parameter_vector_no_vf = np.empty(
                    len(x0_bool_n0_vf), dtype=float)
                parameter_vector_no_vf[~x0_bool_n0_vf] = (
                    optimized_parameter_vector)
                parameter_vector_no_vf[x0_bool_n0_vf] = x0_params[
                    :-self.Nmodels][x0_bool_n0_vf]
                parameter_vector[:-self.Nmodels] = parameter_vector_no_vf
            parameter_vector = (
                parameter_vector * self.model.scales_for_optimization)
            parameters = self.model.parameter_vector_to_parameters(
                parameter_vector)
            phi_x = self.model(acquisition_scheme,
                               quantity="stochastic cost function",
                               **parameters)
            if np.all(~np.isnan(x0_params[-self.Nmodels:])):
                # if initial guess is given for volume fractions
                vf = x0_params[-self.Nmodels:]
            else:
                A = np.dot(phi_x.T, phi_x)
                try:
                    phi_inv = np.dot(np.linalg.inv(A), phi_x.T)
                    vf = np.dot(phi_inv, data)
                except np.linalg.linalg.LinAlgError:
                    # happens when models have the same signal attenuations.
                    vf = np.ones(self.Nmodels) / float(self.Nmodels)
            E_hat = np.dot(phi_x, vf)
        objective = np.dot(data - E_hat, data - E_hat).squeeze()
        return objective * 1e5

    def cobyla_cost_function(self, fractions, phi, data):
        "Objective function of linear parameter estimation using COBYLA."
        E_hat = np.dot(phi, fractions)
        diff = data - E_hat
        objective = np.dot(diff, diff)
        return objective * 1e5

    def objective_function(
            self, optimized_parameter_vector, data, acquisition_scheme,
            x0_params):
        "Objective function of final refining step using L-BFGS-B"
        nested_fractions = optimized_parameter_vector[-(self.Nmodels - 1):]
        normalized_fractions = nested_to_normalized_fractions(
            nested_fractions)
        optimized_parameter_vector = np.r_[
            optimized_parameter_vector[:-(self.Nmodels - 1)],
            normalized_fractions]
        if not np.all(self.model.opt_params_for_optimization):
            x0_bool_array = ~np.isnan(x0_params)
            parameter_vector = np.ones(len(x0_bool_array))
            parameter_vector[~x0_bool_array] = (
                optimized_parameter_vector)
            parameter_vector[x0_bool_array] = x0_params[x0_bool_array]
        else:
            parameter_vector = optimized_parameter_vector

        parameter_vector = (
            parameter_vector * self.model.scales_for_optimization)
        parameters = {}
        parameters.update(
            self.model.parameter_vector_to_parameters(parameter_vector)
        )
        E_model = self.model(acquisition_scheme, **parameters)
        E_diff = E_model - data
        objective = np.dot(E_diff, E_diff) / len(data)
        return objective * 1e5


def nested_to_normalized_fractions(nested_fractions):
    "Function to convert nested to normalized volume fractions."
    N = len(nested_fractions)
    normalized_fractions = np.zeros(N + 1)
    remaining_fraction = 1.
    for i in range(N):
        normalized_fractions[i] = remaining_fraction * nested_fractions[i]
        remaining_fraction -= normalized_fractions[i]
    normalized_fractions[-1] = remaining_fraction
    return normalized_fractions


def cobyla_positivity_constraint(volume_fractions, *args):
    "COBYLA positivity constraint on volume fractions"
    return volume_fractions - 0.001


def cobyla_unity_constraint(volume_fractions, *args):
    "COBYLA unity constraint on volume fractions"
    return np.sum(volume_fractions) - 1




import numpy as np
from ..core.modeling_framework import ModelProperties
from ..core.constants import CONSTANTS

DIAMETER_SCALING = 1e-6


__all__ = [
    'P3PlaneCallaghanApproximation'
]


class P2PlaneStejskalTannerApproximation(ModelProperties):
    r""" Stejskal-Tanner approximation of diffusion between two infinitely
    large parallel planes. Assumes short-gradient pulse (SGP) approximation
    (pulse length towards zero) and the long diffusion time limit (pulse
    separation towards infinity). We follow the notation of Balinov [1]_.

    References
    ----------
    .. [1] Balinov, Balin, et al. "The NMR self-diffusion method applied to
        restricted diffusion. Simulation of echo attenuation from molecules in
        spheres and between planes." Journal of Magnetic Resonance, Series A
        104.1 (1993): 17-25.
    """
    _required_acquisition_parameters = ['qvalues']
    _parameter_ranges = {
        'diameter': (1e-2, 20)
    }

    _parameter_scales = {
        'diameter': DIAMETER_SCALING
    }

    _parameter_types = {
        'diameter': 'plane'
    }
    _model_type = 'NMRModel'

    def __init__(self, diameter=None):
        self.diameter = diameter

    def plane_attenuation(self, q, diameter):
        "Equation 6 in Balinov et al. (1993)."
        q_argument = 2 * np.pi * q * diameter
        return 2 * (1 - np.cos(q_argument)) / q_argument ** 2

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        q = acquisition_scheme.qvalues
        diameter = kwargs.get('diameter', self.diameter)

        E_plane = np.ones_like(q)
        q_nonzero = q > 0
        E_plane[q_nonzero] = self.plane_attenuation(
            q[q_nonzero], diameter
        )
        return E_plane


class P3PlaneCallaghanApproximation(ModelProperties):
    r"""
    The Callaghan model [1]_ of diffusion between two parallel infinite plates.

    Parameters
    ----------
    diameter : float
        Distance between the two plates in meters.
    diffusion_constant : float,
        The diffusion constant of the water particles between the two planes.
        The default value is the approximate diffusivity of water inside axons
        as 1.7e-9 m^2/s.
    number_of_roots : integer,
        The number of roots for the Callaghan approximation.

    References
    ----------
    [1] Callaghan, "Pulsed-Gradient Spin-Echo NMR for Planar, Cylindrical,
        and Spherical Pores under Conditions of Wall Relaxation", JMR 1995
    """
    _required_acquisition_parameters = ['qvalues', 'tau']

    _parameter_ranges = {
        'diameter': (1e-2, 20)
    }

    _parameter_scales = {
        'diameter': DIAMETER_SCALING
    }

    _parameter_types = {
        'diameter': 'plane'
    }
    _model_type = 'NMRModel'

    def __init__(
        self,
        diameter=None,
        diffusion_constant=CONSTANTS['water_in_axons_diffusion_constant'],
        number_of_roots=40,
    ):

        self.diameter = diameter
        self.Dintra = diffusion_constant
        self.xi = np.arange(number_of_roots) * np.pi
        self.zeta = np.arange(number_of_roots) * np.pi + np.pi / 2.0

    def plane_attenuation(self, q, tau, diameter):
        """Implements the finite time Callaghan model for planes."""
        radius = diameter / 2.0
        q_argument = 2 * np.pi * q * radius
        q_argument_2 = q_argument ** 2
        res = np.zeros_like(q)
        for n in range(len(self.xi)):
            xi_n = self.xi[n]
            xi_n2 = self.xi[n] ** 2

            if xi_n == 0.:
                div = 1.
            else:
                div = np.sin(2 * xi_n) / 2 * xi_n

            update = (
                2 * np.exp(-xi_n2 * self.Dintra * tau / radius ** 2) /
                (1 + div) *
                (q_argument * np.sin(q_argument) * np.cos(xi_n) - xi_n *
                 np.cos(q_argument) * np.sin(xi_n)) ** 2 /
                (q_argument_2 - xi_n2) ** 2
            )

            update[~np.isfinite(update)] = 0.

            res += update

        for m in range(len(self.zeta)):
            zeta_m = self.zeta[m]
            zeta_m2 = self.zeta[m] ** 2

            update = (
                2 * np.exp(-zeta_m2 * self.Dintra * tau / radius ** 2) /
                (1 - np.sin(2 * zeta_m) / (2 * zeta_m)) *
                (q_argument * np.cos(q_argument) * np.sin(zeta_m) - zeta_m *
                 np.sin(q_argument) * np.cos(zeta_m)) ** 2 /
                (q_argument_2 - zeta_m2) ** 2
            )

            update[~np.isfinite(update)] = 0.
            res += update
        return res

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        q = acquisition_scheme.qvalues
        tau = acquisition_scheme.tau
        diameter = kwargs.get('diameter', self.diameter)

        E_plane = np.ones_like(q)
        q_nonzero = q > 0
        E_plane[q_nonzero] = self.plane_attenuation(
            q[q_nonzero], tau[q_nonzero], diameter
        )
        return E_plane



import numpy as np
from dmipy.utils.utils import cart2mu
from dmipy.utils.spherical_convolution import real_sym_rh_basis
from ..core.modeling_framework import ModelProperties
from ..core.signal_model_properties import (
    AnisotropicSignalModelProperties,
    IsotropicSignalModelProperties)
from ..utils import utils
from dmipy.core.acquisition_scheme import gtab_dmipy2dipy
from dipy.reconst import dti


def estimate_TR1_isotropic_tissue_response_model(acquisition_scheme, data):
    """
    Estimates isotropic TissueResponseModel describing the convolution
    kernel of e.g. CSF or grey matter from array of candidate voxels [1]_.

    First, for each acquisition shell, the zeroth order rotational harmonic
    (RH) coefficient (so actually only the Y00 coefficient) is estimated.

    From the signal RH coefficients the S0 response and the signal attenuation
    RH coefficients are then separated. The TR2 model is then instantiated with
    these RH coefficients and returns allong with the S0 response.

    Parameters
    ----------
    acquisition_scheme : DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using dMipy.
    data : 2D array of size (N_voxels, N_DWIs),
            Candidate diffusion signal array to generate anisotropic tissue
            response from.

    Returns
    -------
    S0_response: positive float,
        average positive S0 response value of the input data.
    TR1: Dmipy TR1IsotropicTissueResponse instance,
        average shape model of the input data, defined only on the input data
        shells. Can be used as usual CompartmentModel.

    References
    ----------
    .. [1] Tournier, J-Donald, Fernando Calamante, and Alan Connelly.
        "Determination of the appropriate b-value and number of gradient
        directions for high-angular-resolution diffusion-weighted imaging."
        NMR in Biomedicine 26.12 (2013): 1775-1786.
    """

    if acquisition_scheme.N_TE > 1:
        msg = 'tissue response estimation not yet implemented for data with '\
              'multiple TE. Current data has {} TEs.'
        raise NotImplementedError(msg.format(acquisition_scheme.N_TE))

    rh_matrices = np.zeros((len(data), acquisition_scheme.N_dwi_shells, 1))
    S0_response = np.mean(data[:, acquisition_scheme.b0_mask])

    inv_sh_mats = {}
    for shell_index, sh_mat in acquisition_scheme.shell_sh_matrices.items():
        inv_sh_mats[shell_index] = np.linalg.pinv(sh_mat)

    for i in range(len(data)):
        for j, shell_index in enumerate(acquisition_scheme.unique_dwi_indices):
            shell_mask = acquisition_scheme.shell_indices == shell_index
            shell_data = data[i][shell_mask]
            rh_matrices[i, j, 0] = np.dot(
                inv_sh_mats[shell_index], shell_data)[0]
    rotational_harmonics_representation = np.mean(
        rh_matrices, axis=0) / S0_response
    TR1 = TR1IsotropicTissueResponseModel(
        acquisition_scheme, rotational_harmonics_representation)
    return S0_response, TR1


def estimate_TR2_anisotropic_tissue_response_model(acquisition_scheme, data):
    """
    Estimates TR2 anistropic TissueResponseModel describing the convolution
    kernel of e.g. anistropic white matter from array of candidate voxels [1]_

    First, Each candidate voxel is rotated such that the DTI eigenvector with
    the largest eigenvalue is aligned with the z-axis. The rotational harmonic
    (RH) coefficients (corresponding to Y_l0 spherical harmonics) are then
    estimated and saved per acquisition shell.

    From the signal RH coefficients the S0 response and the signal attenuation
    RH coefficients are then separated. The TR2 model is then instantiated with
    these RH coefficients and returns allong with the S0 response.

    Parameters
    ----------
    acquisition_scheme : DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using Dmipy.
    data : 2D array of size (N_voxels, N_DWIs),
            Candidate diffusion signal array to generate anisotropic tissue
            response from.

    Returns
    -------
    S0_response: positive float,
        average positive S0 response value of the input data.
    TR2: Dmipy TR2AnisotropicTissueResponse instance,
        average shape model of the input data, defined only on the input data
        shells. Can be used as usual CompartmentModel.

    References
    ----------
    .. [1] Tournier, J-Donald, Fernando Calamante, and Alan Connelly.
        "Determination of the appropriate b-value and number of gradient
        directions for high-angular-resolution diffusion-weighted imaging."
        NMR in Biomedicine 26.12 (2013): 1775-1786.
    """
    if acquisition_scheme.N_TE > 1:
        msg = 'tissue response estimation not yet implemented for data with '\
              'multiple TE. Current data has {} TEs.'
        raise NotImplementedError(msg.format(acquisition_scheme.N_TE))
    gtab = gtab_dmipy2dipy(acquisition_scheme)
    tenmod = dti.TensorModel(gtab)
    tenfit = tenmod.fit(data)
    evecs = tenfit.evecs
    max_sh_order = max(acquisition_scheme.shell_sh_orders.values())
    rh_matrices = np.zeros(
        (len(data),
         acquisition_scheme.N_dwi_shells, int(max_sh_order // 2 + 1)))
    S0_response = np.mean(data[:, acquisition_scheme.b0_mask])

    for i in range(len(data)):
        # dipy's evecs are automatically ordered such that
        # lambda1 > lambda2 > lambda3 in xyz coordinate system. To estimate
        # rotational harmonics we need the eigenvector corresponding to the
        # largest lambda1 to be along the z-axis. This is why we rotate
        # the gradient directions with the reverse of the dti eigenvectors.
        bvecs_rot = np.dot(acquisition_scheme.gradient_directions,
                           evecs[i][:, ::-1])
        for j, shell_index in enumerate(acquisition_scheme.unique_dwi_indices):
            shell_sh = acquisition_scheme.shell_sh_orders[shell_index]
            shell_mask = acquisition_scheme.shell_indices == shell_index
            shell_bvecs = bvecs_rot[shell_mask]
            theta, phi = cart2mu(shell_bvecs).T
            rh_mat = real_sym_rh_basis(shell_sh, theta, phi)
            rh_matrices[i, j, :shell_sh // 2 + 1] = np.dot(
                np.linalg.pinv(rh_mat), data[i][shell_mask])
    rotational_harmonics_representation = np.mean(
        rh_matrices, axis=0) / S0_response

    TR2 = TR2AnisotropicTissueResponseModel(
        acquisition_scheme, rotational_harmonics_representation)

    return S0_response, TR2


class TR1IsotropicTissueResponseModel(
        ModelProperties, IsotropicSignalModelProperties):
    """
    The isotropic tissue response model is a non-parametric multi-shell
    representation of some signal attenuation. It can be instantiated with any
    rotational harmonics and accompanying DmipyAcquisitionScheme, but is
    usually instantiated from a set of segmented input data using the
    estimate_TR1_anisotropic_tissue_response_model function.

    Once estimated, this class behaves as a CompartmentModel object (so as if
    it were e.g. a cylinder or Gaussian compartment), but has no parameters.
    A TissueResponseModel can be input any MultiCompartment model
    representation in Dmipy, including together with parametric models.

    NOTE: TR models can ONLY generate signal attenuation at the same shells as
    the input rotational harmonics. However, TR models can be called with
    partial acquisition schemes. Meaning, ifthe TR model was instantiated with
    multi-shell acquisition scheme, and there is another scheme which is a
    subset of that scheme, it can be used to generate the partial data.

    Parameters
    ----------
    acquisition_scheme: DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using Dmipy.
    rotational_harmonics: array, shape(Nshells, 1),
            Y00 rotational harmonics coefficients for each shell.
    """
    _required_acquisition_parameters = []
    _parameter_ranges = {}
    _parameter_scales = {}
    _parameter_types = {}
    _model_type = 'TissueResponseModel'

    def __init__(self, acquisition_scheme, rotational_harmonics):
        self.acquisition_scheme = acquisition_scheme
        self._rotational_harmonics = rotational_harmonics

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Returns the signal attenuation. As a tissue response model, it cannot
        estimate any signal intensities outside the shells that were in the
        acquisition scheme that was used to estimate the model. However, it can
        estimate any gradient directions on the estimated shells.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E : float or array, shape(N),
            signal attenuation
        '''
        E = np.ones(acquisition_scheme.number_of_measurements, dtype=float)
        E_mean = self.spherical_mean(acquisition_scheme)
        for shell_index in acquisition_scheme.unique_dwi_indices:
            shell_mask = acquisition_scheme.shell_indices == shell_index
            E[shell_mask] = E_mean[shell_index]
        return E

    def rotational_harmonics_representation(
            self, acquisition_scheme, **kwargs):
        r""" The rotational harmonics of the model, such that Y_lm = Yl0.
        Axis aligned with z-axis to be used as kernel for spherical
        convolution. Returns an array with rotational harmonics for each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        rh_array : array, shape(Nshells, 1),
            Y00 rotational harmonics coefficients for each shell.
        """
        rh_scheme = acquisition_scheme.rotational_harmonics_scheme
        rh_array = np.zeros((rh_scheme.N_dwi_shells, 1))
        for i, dwi_shell_index in enumerate(rh_scheme.unique_dwi_indices):
            b_value = acquisition_scheme.shell_bvalues[dwi_shell_index]
            shell_overlap = self.acquisition_scheme.shell_bvalues[
                ~self.acquisition_scheme.shell_b0_mask] == b_value
            if not np.any(shell_overlap):
                raise ValueError(
                    'bvalue {}s/mm^2 not in tissue response bvalues'.format(
                        b_value / 1e6))
            native_dwi_shell_index = np.where(shell_overlap)[0][0]
            rh_array[i] = self._rotational_harmonics[native_dwi_shell_index]
        return rh_array

    def spherical_mean(self, acquisition_scheme, **kwargs):
        """
        Estimates spherical mean for every shell in acquisition scheme for
        Zeppelin model.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E_mean : float,
            spherical mean of the Zeppelin model for every acquisition shell.
        """
        E_mean = np.ones(acquisition_scheme.N_shells)
        rh_coef = self.rotational_harmonics_representation(acquisition_scheme)
        E_mean[~acquisition_scheme.shell_b0_mask] = (
            rh_coef[:, 0] / (2 * np.sqrt(np.pi)))
        return E_mean


class TR2AnisotropicTissueResponseModel(
        ModelProperties, AnisotropicSignalModelProperties):
    """
    The anisotropic tissue response model is a non-parametric multi-shell
    representation of some signal attenuation. It can be instantiated with any
    rotational harmonics and accompanying DmipyAcquisitionScheme, but is
    usually instantiated from a set of segmented input data using the
    estimate_TR2_anisotropic_tissue_response_model function.

    Once instantiated, this class behaves as a CompartmentModel object (so as
    if it were e.g. a cylinder or Gaussian compartment), but only has an
    orientation parameter. A TissueResponseModel can be input any
    MultiCompartment model representation in Dmipy, including together with
    parametric models.

    NOTE: TR models can ONLY generate signal attenuation at the same shells as
    the input rotational harmonics. However, TR models can be called with
    partial acquisition schemes. Meaning, ifthe TR model was instantiated with
    multi-shell acquisition scheme, and there is another scheme which is a
    subset of that scheme, it can be used to generate the partial data.

    Parameters
    ----------
    acquisition_scheme : DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using Dmipy.
    rotational_harmonics: array of shape (N_shell_DWI, N_rh_coeff),
        Rotational harmonics that define the shape of the TR model.
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    """
    _required_acquisition_parameters = []
    _parameter_ranges = {'mu': ([0, np.pi], [-np.pi, np.pi])}
    _parameter_scales = {'mu': np.r_[1., 1.]}
    _parameter_types = {'mu': 'orientation'}
    _model_type = 'TissueResponseModel'

    def __init__(self, acquisition_scheme, rotational_harmonics, mu=None):
        self.acquisition_scheme = acquisition_scheme
        self._rotational_harmonics = rotational_harmonics
        self.mu = mu

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Returns the signal attenuation. As a tissue response model, it cannot
        estimate any signal intensities outside the shells that were in the
        acquisition scheme that was used to estimate the model. However, it can
        estimate any gradient directions on the estimated shells as it just
        maps the rotational harmonics to the shell for each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E : float or array, shape(N),
            signal attenuation
        '''
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)

        # rotate gradient vectors according to orientation parameters
        R = utils.rotation_matrix_001_to_xyz(*mu)
        bvecs = acquisition_scheme.gradient_directions
        bvecs_rot = np.dot(bvecs, R)

        # now map the rotational harmonics to the rotated gradient vectors.
        rh_coef = self.rotational_harmonics_representation(acquisition_scheme)
        E = np.ones(acquisition_scheme.number_of_measurements, dtype=float)
        scheme = acquisition_scheme
        for i, dwi_shell_index in enumerate(scheme.unique_dwi_indices):
            shell_sh = scheme.shell_sh_orders[dwi_shell_index]
            shell_mask = scheme.shell_indices == dwi_shell_index
            shell_bvecs = bvecs_rot[shell_mask]
            theta, phi = cart2mu(shell_bvecs).T
            rh_mat = real_sym_rh_basis(shell_sh, theta, phi)
            E[shell_mask] = np.dot(rh_mat, rh_coef[i, :shell_sh // 2 + 1])
        return E

    def rotational_harmonics_representation(
            self, acquisition_scheme, **kwargs):
        r""" The rotational harmonics of the model, such that Y_lm = Yl0.
        Axis aligned with z-axis to be used as kernel for spherical
        convolution. Returns an array with rotational harmonics for each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        rh_array : array, shape(Nshells, N_rh_coef),
            Rotational harmonics coefficients for each shell.
        """
        rh_scheme = acquisition_scheme.rotational_harmonics_scheme
        max_sh_order = max(rh_scheme.shell_sh_orders.values())
        rh_array = np.zeros((rh_scheme.N_dwi_shells, max_sh_order // 2 + 1))
        for i, dwi_shell_index in enumerate(rh_scheme.unique_dwi_indices):
            b_value = acquisition_scheme.shell_bvalues[dwi_shell_index]
            shell_overlap = self.acquisition_scheme.shell_bvalues[
                ~self.acquisition_scheme.shell_b0_mask] == b_value
            if not np.any(shell_overlap):
                raise ValueError(
                    'bvalue {}s/mm^2 not in tissue response bvalues'.format(
                        b_value / 1e6))
            native_dwi_shell_index = np.where(shell_overlap)[0][0]
            rh_array[i] = self._rotational_harmonics[
                native_dwi_shell_index, :max_sh_order // 2 + 1]
        return rh_array

    def spherical_mean(self, acquisition_scheme, **kwargs):
        """
        Estimates spherical mean for every shell in acquisition scheme for
        Zeppelin model.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E_mean : float,
            spherical mean of the Zeppelin model for every acquisition shell.
        """
        E_mean = np.ones(acquisition_scheme.N_shells)
        rh_coef = self.rotational_harmonics_representation(acquisition_scheme)
        E_mean[~acquisition_scheme.shell_b0_mask] = (
            rh_coef[:, 0] / (2 * np.sqrt(np.pi)))
        return E_mean



import numpy as np
from dipy.data import get_sphere, HemiSphere
from dipy.reconst.shm import real_sym_sh_mrtrix
from dipy.utils.optpkg import optional_package
from dipy.reconst.shm import sph_harm_ind_list
sphere = get_sphere('symmetric724')
numba, have_numba, _ = optional_package("numba")


__all__ = [
    'CsdTournierOptimizer'
]


class CsdTournierOptimizer:
    """
    The classical Constrained Spherical Deconvolution (CSD) optimizer as
    proposed by Tournier et al. (2007) [1]_.

    It is a bit less accurate than the general CVXPY optimizer, but
    significantly faster.

    TODO: multicore processing makes this solver MUCH slower for some
    reason. It is likely the fact that it's copying the kernel and
    positivity matrices to every process.

    Parameters
    ----------
    acquisition_scheme: DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using Dmipy.
    model: Dmipy MultiCompartmentSphericalHarmonicsModel instance,
        Contains Dmipy multi-compartment model information.
    x0_vector: Array of size (Nparameters)
        Possible parameters for model kernels.
    sh_order: positive even integer,
        Spherical harmonics order for deconvolution.
    lambda_pos: positive float,
        Positivity regularization parameter.
    lambda_lb: positive float,
        Laplace-Belrami regularization weight to impose smoothness in the
        FOD. Same as is done in [2]_.
    tau: positive float,
        Scales positivity threshold relative to maximum FOD amplitude.
    max_iter: positive integer,
        Maximum number of iterations for optimization.
    unity_constraint: boolean,
        whether or not to constrain the volume fractions of the FOD to
        unity. In the case of one model this means that the SH-coefficients
        represent a distribution on the sphere.
    init_sh_order: positive even integer,
        Spherical harmonics order used to calculate unconstrained initial
        guess for the FOD coefficients.

    References
    ----------
    .. [1] Tournier, J-Donald, Fernando Calamante, and Alan Connelly.
        "Robust determination of the fibre orientation distribution in
        diffusion MRI: non-negativity constrained super-resolved spherical
        deconvolution." Neuroimage 35.4 (2007): 1459-1472.
    .. [2] Descoteaux, Maxime, et al. "Regularized, fast, and robust analytical
        Q-ball imaging." Magnetic Resonance in Medicine: An Official Journal of
        the International Society for Magnetic Resonance in Medicine 58.3
        (2007): 497-510.
    """

    def __init__(self, acquisition_scheme, model, x0_vector=None, sh_order=8,
                 lambda_pos=1., lambda_lb=5e-4, tau=0.1, max_iter=50,
                 unity_constraint=True, init_sh_order=4):
        self.model = model
        self.acquisition_scheme = acquisition_scheme
        self.sh_order = sh_order
        self.Ncoef = int((sh_order + 2) * (sh_order + 1) // 2)
        self.Ncoef4 = int((init_sh_order + 2) * (init_sh_order + 1) // 2)
        self.Nmodels = len(self.model.models)
        self.lambda_pos = lambda_pos
        self.lambda_lb = lambda_lb
        self.tau = tau
        self.max_iter = max_iter
        self.unity_constraint = unity_constraint
        self.sphere_jacobian = 1 / (2 * np.sqrt(np.pi))

        # step 1: prepare positivity grid on sphere
        sphere = get_sphere('symmetric724')
        hemisphere = HemiSphere(phi=sphere.phi, theta=sphere.theta)
        self.L_positivity = real_sym_sh_mrtrix(
            self.sh_order, hemisphere.theta, hemisphere.phi)[0]

        sh_l = sph_harm_ind_list(sh_order)[1]
        self.R_smoothness = np.diag(sh_l ** 2 * (sh_l + 1) ** 2)

        # check if there is only one model. If so, precompute rh array.
        if self.model.volume_fractions_fixed:
            x0_single_voxel = np.reshape(
                x0_vector, (-1, x0_vector.shape[-1]))[0]
            if np.all(np.isnan(x0_single_voxel)):
                self.single_convolution_kernel = True
                parameters_dict = self.model.parameter_vector_to_parameters(
                    x0_single_voxel)
                self.A = self.model._construct_convolution_kernel(
                    **parameters_dict)
                self.AT_A = np.dot(self.A.T, self.A)
            else:
                self.single_convolution_kernel = False
        else:
            msg = "This CSD optimizer cannot estimate volume fractions."
            raise ValueError(msg)

    def __call__(self, data, x0_vector):
        """
        The fitting function of Tournier's CSD optimizer.

        If there is only one convolution kernel, it loads the precalculated
        values. If the kernel is voxel-varying, it calculates it now.

        Calls a different optimizing procedure depending on if
        the unity_constraint is True or False.

        Parameters
        ----------
        data : Array of size (Ndata)
            The normalized dMRI signal attenuation to be fitted.
        x0_vector : Array of size (Nparameters)
            Possible initial guess vector for parameter initiation.

        Returns
        -------
        fitted_parameter_vector : Array of size (Nparameters),
            The fitted MC model parameters using CSD.
        """

        if self.single_convolution_kernel:
            A = self.A
            AT_A = self.AT_A + self.lambda_lb * self.R_smoothness
        else:
            parameters_dict = self.model.parameter_vector_to_parameters(
                x0_vector)
            A = self.model._construct_convolution_kernel(**parameters_dict)
            AT_A = np.dot(A.T, A) + self.lambda_lb * self.R_smoothness

        if self.unity_constraint:
            return self._optimize_with_unity_constraint(
                A, AT_A, data, x0_vector)
        else:
            return self._optimize_without_unity_constraint(
                A, AT_A, data, x0_vector)

    def _optimize_without_unity_constraint(self, A, AT_A, data, x0_vector):
        """
        CSD optimizer that does not take into account the unity constraint of
        the FOD distribution.

        Parameters
        ----------
        A: array of size (N_data, N_coef),
            Observation matrix that maps FOD coefficients to DWI intensities.
        AT_A: array of size (N_coef, Ncoef),
            Precomputation of np.dot(A.T, A)
        data: array of size (N_data,)
            Measured DWI data.
        x0_vector: array of size (N_parameters),
            Initial condition for parameters.

        Returns
        -------
        fitted_parameter_vector: array of size (N_parameters),
            Fitted parameters including SH coefficients of FOD.
        """
        f_sh = np.zeros(self.Ncoef, dtype=float)

        # initialize coefficients with initial guess
        f_sh[:self.Ncoef4] = np.dot(np.linalg.pinv(A[:, :self.Ncoef4]), data)

        # minimum threshold of FOD amplitude based on FOD mean (without
        # jacobian) according to dipy.
        threshold = self.tau * self.L_positivity[0, 0] * f_sh[0]
        negative_fod_check = np.dot(self.L_positivity, f_sh) < threshold

        for iteration in range(self.max_iter):
            L = self.L_positivity[negative_fod_check]
            Q = AT_A + self.lambda_pos * np.dot(L.T, L)
            try:
                f_sh = np.dot(np.dot(np.linalg.inv(Q), A.T), data)
            except np.linalg.LinAlgError:
                Q += 1e-5 * np.eye(Q.shape[0])
            negative_fod_check_old = negative_fod_check
            negative_fod_check = np.dot(self.L_positivity, f_sh) < threshold
            if np.array_equal(negative_fod_check, negative_fod_check_old):
                break
        # return optimized fod sh coefficients
        fitted_params = self.model.parameter_vector_to_parameters(x0_vector)
        fitted_params['sh_coeff'] = f_sh
        fitted_parameter_vector = self.model.parameters_to_parameter_vector(
            **fitted_params)
        return fitted_parameter_vector

    def _optimize_with_unity_constraint(self, A, AT_A, data, x0_vector):
        """
        CSD optimizer that constrains the first coefficient to be equal to
        1 / (2 * np.sqrt(np.pi)), in order for the spherical integration of
        the FOD to add up to unity.

        In the current implementation the first coefficient is just set to
        this value and never updated, only allowing other to change in the
        iteration.

        A better approach is to precalculate the remained of the data by
        subtracting the convolution of the first known coefficient and the
        convolution kernel from the data. Then only fitting the remaining
        coefficients to the remaining data should result in a better
        estimate of the coefficients. However, the positivity regularization
        seems to not behave the same way when doing this, which is why this
        is the current solution.

        Parameters
        ----------
        A: array of size (N_data, N_coef),
            Observation matrix that maps FOD coefficients to DWI intensities.
        AT_A: array of size (N_coef, Ncoef),
            Precomputation of np.dot(A.T, A)
        data: array of size (N_data,)
            Measured DWI data.
        x0_vector: array of size (N_parameters),
            Initial condition for parameters.

        Returns
        -------
        fitted_parameter_vector: array of size (N_parameters),
            Fitted parameters including SH coefficients of FOD.
        """
        f_sh = np.zeros(self.Ncoef, dtype=float)
        f_sh[0] = self.sphere_jacobian

        # initialize coefficients with initial guess
        f_sh[1:self.Ncoef4] = np.dot(
            np.linalg.pinv(A[:, :self.Ncoef4]), data)[1:]

        # minimum threshold of FOD amplitude based on FOD mean (without
        # jacobian) according to dipy.
        threshold = self.tau * self.L_positivity[0, 0] * f_sh[0]
        negative_fod_check = np.dot(self.L_positivity, f_sh) < threshold

        for iteration in range(self.max_iter):
            L = self.L_positivity[negative_fod_check]
            Q = AT_A + self.lambda_pos * np.dot(L.T, L)
            f_sh[1:] = np.dot(np.dot(np.linalg.inv(Q), A.T), data)[1:]
            negative_fod_check_old = negative_fod_check
            negative_fod_check = np.dot(self.L_positivity, f_sh) < threshold
            if np.array_equal(negative_fod_check, negative_fod_check_old):
                break
        # return optimized fod sh coefficients
        fitted_params = self.model.parameter_vector_to_parameters(x0_vector)
        fitted_params['sh_coeff'] = f_sh
        fitted_parameter_vector = self.model.parameters_to_parameter_vector(
            **fitted_params)
        return fitted_parameter_vector



import numpy as np
from dipy.data import get_sphere
from dipy.reconst.shm import sh_to_sf_matrix, real_sym_sh_mrtrix
from ..utils.utils import (
    unitsphere2cart_Nd, T1_tortuosity, fractional_parameter, cart2mu)
from ..utils.spherical_mean import (
    estimate_spherical_mean_multi_shell)
from dipy.reconst.shm import sph_harm_ind_list
from dipy.direction import peak_directions


__all__ = [
    'FittedMultiCompartmentModel',
    'FittedMultiCompartmentSphericalMeanModel'
]


class FittedMultiCompartmentModel:
    """
    The FittedMultiCompartmentModel instance contains information about the
    original MultiCompartmentModel, the estimated S0 values, the fitting mask
    and the fitted model parameters.

    Parameters
    ----------
    model : MultiCompartmentModel instance,
        A dmipy MultiCompartmentModel.
    S0 : array of size (Ndata,) or (N_data, N_DWIs),
        Array containing the estimated S0 values of the data. If data is 4D,
        then S0 is 3D if there is only one TE, and the same 4D size of the data
        if there are multiple TEs.
    mask : array of size (N_data,),
        boolean mask of voxels that were fitted.
    fitted_parameters_vector : array of size (N_data, Nparameters),
        fitted model parameters array.
    """

    def __init__(self, model, S0, mask, fitted_parameters_vector,
                 fitted_multi_tissue_fractions_vector=None):
        self.model = model
        self.S0 = S0
        self.mask = mask
        self.fitted_parameters_vector = fitted_parameters_vector
        self.fitted_multi_tissue_fractions_vector = (
            fitted_multi_tissue_fractions_vector)

    @property
    def fitted_parameters(self):
        "Returns the fitted parameters as a dictionary."
        return self.model.parameter_vector_to_parameters(
            self.fitted_parameters_vector)

    @property
    def fitted_multi_tissue_fractions(self):
        "Returns the fitted multi tissue fractions as a dictionary."
        return self._return_fitted_multi_tissue_fractions()

    @property
    def fitted_multi_tissue_fractions_normalized(self):
        "Returns the normalized fitted multi tissue fractions as a dictionary"
        return self._return_fitted_multi_tissue_fractions(normalized=True)

    def _return_fitted_multi_tissue_fractions(self, normalized=False):
        """
        Returns the multi-tissue estimated volume fractions.

        Parameters
        ----------
        normalized: boolean,
            whether or not to normalize returned multi-tissue volume fractions.
            NOTE: This does not mean the unity constraint was enforced during
            the estimation of the fractions - it is just a normalization after
            the fact.

        Returns
        -------
        mt_partial_volumes: dict,
            contains the multi-tissue volume fractions by name.
            NOTE: if the MC-model only consisted of 1 model, then the name will
            be 'partial_volume_0', but will not have a counterpart in
            self.fitted_parameters.
        """
        mt_fract_vec = self.fitted_multi_tissue_fractions_vector.copy()
        if normalized:
            mt_fract_sum = np.sum(mt_fract_vec, axis=-1)
            mt_fract_mask = mt_fract_sum > 0
            mt_fract_vec[mt_fract_mask] = (
                mt_fract_vec[mt_fract_mask] /
                mt_fract_sum[mt_fract_mask][..., None])
        if self.model.N_models > 1:
            fract_names = self.model.partial_volume_names
        else:
            fract_names = ['partial_volume_0']
        mt_partial_volumes = {}
        for i, partial_volume_name in enumerate(fract_names):
            mt_partial_volumes[partial_volume_name] = mt_fract_vec[..., i]
        return mt_partial_volumes

    @property
    def fitted_and_linked_parameters(self):
        "Returns the fitted and linked parameters as a dictionary."
        fitted_parameters = self.model.parameter_vector_to_parameters(
            self.fitted_parameters_vector)
        return self.model.add_linked_parameters_to_parameters(
            fitted_parameters)

    def fod(self, vertices, visual_odi_lower_bound=0.):
        """
        Returns the Fiber Orientation Distribution if it is available.

        Parameters
        ----------
        vertices : array of size (Nvertices, 3),
            Array of cartesian unit vectors at which to sample the FOD.
        visual_odi_lower_bound : float,
            gives a lower bound to the Orientation Distribution Index (ODI) of
            FODs of Watson and Bingham distributions. This can be useful to
            visualize FOD fields where some FODs are extremely sharp.

        Returns
        -------
        fods : array of size (Ndata, Nvertices),
            the FODs of the fitted model, scaled by volume fraction.
        """
        if not self.model.fod_available:
            msg = ('FODs not available for current model.')
            raise ValueError(msg)
        dataset_shape = self.fitted_parameters_vector.shape[:-1]
        N_samples = len(vertices)
        fods = np.zeros(np.r_[dataset_shape, N_samples])
        mask_pos = np.where(self.mask)
        for pos in zip(*mask_pos):
            parameters = self.model.parameter_vector_to_parameters(
                self.fitted_parameters_vector[pos])
            if visual_odi_lower_bound > 0:
                param_keys = parameters.keys()
                for key in param_keys:
                    if key[-3:] == 'odi':
                        parameters[key] = np.clip(parameters[key],
                                                  visual_odi_lower_bound, 1)
            fods[pos] = self.model(vertices, quantity='FOD', **parameters)
        return fods

    def fod_sh(self, sh_order=8, basis_type=None):
        """
        Returns the spherical harmonics coefficients of the Fiber Orientation
        Distribution (FOD) if it is available. Uses are 724 spherical
        tessellation to do the spherical harmonics transform.

        Parameters
        ----------
        sh_order : integer,
            the maximum spherical harmonics order of the coefficient expansion.
        basis_type : string,
            type of spherical harmonics basis to use for the expansion, see
            sh_to_sf_matrix for more info.

        Returns
        -------
        fods_sh : array of size (Ndata, Ncoefficients),
            spherical harmonics coefficients of the FODs, scaled by volume
            fraction.
        """
        if not self.model.fod_available:
            msg = ('FODs not available for current model.')
            raise ValueError(msg)
        sphere = get_sphere(name='repulsion724')
        vertices = sphere.vertices
        _, inv_sh_matrix = sh_to_sf_matrix(
            sphere, sh_order, basis_type=basis_type, return_inv=True)
        fods_sf = self.fod(vertices)

        dataset_shape = self.fitted_parameters_vector.shape[:-1]
        number_coef_used = int((sh_order + 2) * (sh_order + 1) // 2)
        fods_sh = np.zeros(np.r_[dataset_shape, number_coef_used])
        mask_pos = np.where(self.mask)
        for pos in zip(*mask_pos):
            fods_sh[pos] = np.dot(inv_sh_matrix.T, fods_sf[pos])
        return fods_sh

    def peaks_spherical(self):
        "Returns the peak angles of the model."
        mu_params = []
        for name, card in self.model.parameter_cardinality.items():
            if name[-2:] == 'mu' and card == 2:
                mu_params.append(self.fitted_parameters[name])
        if len(mu_params) == 0:
            msg = ('peaks not available for current model.')
            raise ValueError(msg)
        if len(mu_params) == 1:
            return np.expand_dims(mu_params[0], axis=-2)
        return np.concatenate([mu[..., None] for mu in mu_params], axis=-1)

    def peaks_cartesian(self):
        "Returns the cartesian peak unit vectors of the model."
        peaks_spherical = self.peaks_spherical()
        peaks_cartesian = unitsphere2cart_Nd(peaks_spherical)
        return peaks_cartesian

    def predict(self, acquisition_scheme=None, S0=None, mask=None):
        """
        simulates the dMRI signal of the fitted MultiCompartmentModel for the
        estimated model parameters. If no acquisition_scheme is given, then
        the same acquisition_scheme that was used for the fitting is used. If
        no S0 is given then it is assumed to be the estimated one. If no mask
        is given then all voxels are assumed to have been fitted.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        S0 : None or float,
            Signal intensity without diffusion sensitization. If None, uses
            estimated SO from fitting process. If float, uses that value.
        mask : (N-1)-dimensional integer/boolean array of size (N_x, N_y, ...),
            mask of voxels to simulate data at.

        Returns
        -------
        predicted_signal : array of size (Ndata, N_DWIS),
            predicted DWIs for the given model parameters and acquisition
            scheme.
        """
        if acquisition_scheme is None:
            acquisition_scheme = self.model.scheme
        self.model._check_model_params_with_acquisition_params(
            acquisition_scheme)

        dataset_shape = self.fitted_parameters_vector.shape[:-1]
        if S0 is None:
            S0 = self.S0
        elif isinstance(S0, float):
            S0 = np.ones(dataset_shape) * S0
        if mask is None:
            mask = self.mask

        N_samples = len(acquisition_scheme.bvalues)

        predicted_signal = np.zeros(np.r_[dataset_shape, N_samples])
        mask_pos = np.where(mask)
        for pos in zip(*mask_pos):
            parameters = self.model.parameter_vector_to_parameters(
                self.fitted_parameters_vector[pos])
            predicted_signal[pos] = self.model(
                acquisition_scheme, **parameters) * S0[pos]
        return predicted_signal

    def R2_coefficient_of_determination(self, data):
        "Calculates the R-squared of the model fit."
        data_ = data / self.S0[..., None]

        y_hat = self.predict(S0=1.)
        y_bar = np.mean(data_, axis=-1)
        SStot = np.sum((data_ - y_bar[..., None]) ** 2, axis=-1)
        SSres = np.sum((data_ - y_hat) ** 2, axis=-1)
        R2 = 1 - SSres / SStot
        R2[~self.mask] = 0
        return R2

    def mean_squared_error(self, data):
        "Calculates the mean squared error of the model fit."
        if self.model.scheme.TE is None:
            data_ = data / self.S0[..., None]
        else:
            data_ = data / self.S0

        y_hat = self.predict(S0=1.)
        mse = np.mean((data_ - y_hat) ** 2, axis=-1)
        mse[~self.mask] = 0
        return mse


class FittedMultiCompartmentSphericalMeanModel:
    """
    The FittedMultiCompartmentModel instance contains information about the
    original MultiCompartmentModel, the estimated S0 values, the fitting mask
    and the fitted model parameters.

    Parameters
    ----------
    model : MultiCompartmentModel instance,
        A dmipy MultiCompartmentModel.
    S0 : array of size (Ndata,) or (N_data, N_DWIs),
        Array containing the estimated S0 values of the data. If data is 4D,
        then S0 is 3D if there is only one TE, and the same 4D size of the data
        if there are multiple TEs.
    mask : array of size (N_data,),
        boolean mask of voxels that were fitted.
    fitted_parameters_vector : array of size (N_data, Nparameters),
        fitted model parameters array.
    """

    def __init__(self, model, S0, mask, fitted_parameters_vector,
                 fitted_multi_tissue_fractions_vector=None):
        self.model = model
        self.S0 = S0
        self.mask = mask
        self.fitted_parameters_vector = fitted_parameters_vector
        self.fitted_multi_tissue_fractions_vector = (
            fitted_multi_tissue_fractions_vector)

    @property
    def fitted_parameters(self):
        "Returns the fitted parameters as a dictionary."
        return self.model.parameter_vector_to_parameters(
            self.fitted_parameters_vector)

    @property
    def fitted_and_linked_parameters(self):
        "Returns the fitted and linked parameters as a dictionary."
        fitted_parameters = self.model.parameter_vector_to_parameters(
            self.fitted_parameters_vector)
        return self.model.add_linked_parameters_to_parameters(
            fitted_parameters)

    @property
    def fitted_multi_tissue_fractions(self):
        "Returns the fitted multi tissue fractions as a dictionary."
        return self._return_fitted_multi_tissue_fractions()

    @property
    def fitted_multi_tissue_fractions_normalized(self):
        "Returns the normalized fitted multi tissue fractions as a dictionary"
        return self._return_fitted_multi_tissue_fractions(normalized=True)

    def _return_fitted_multi_tissue_fractions(self, normalized=False):
        """
        Returns the multi-tissue estimated volume fractions.

        Parameters
        ----------
        normalized: boolean,
            whether or not to normalize returned multi-tissue volume fractions.
            NOTE: This does not mean the unity constraint was enforced during
            the estimation of the fractions - it is just a normalization after
            the fact.

        Returns
        -------
        mt_partial_volumes: dict,
            contains the multi-tissue volume fractions by name.
            NOTE: if the MC-model only consisted of 1 model, then the name will
            be 'partial_volume_0', but will not have a counterpart in
            self.fitted_parameters.
        """
        mt_fract_vec = self.fitted_multi_tissue_fractions_vector.copy()
        if normalized:
            mt_fract_sum = np.sum(mt_fract_vec, axis=-1)
            mt_fract_mask = mt_fract_sum > 0
            mt_fract_vec[mt_fract_mask] = (
                mt_fract_vec[mt_fract_mask] /
                mt_fract_sum[mt_fract_mask][..., None])
        if self.model.N_models > 1:
            fract_names = self.model.partial_volume_names
        else:
            fract_names = ['partial_volume_0']
        mt_partial_volumes = {}
        for i, partial_volume_name in enumerate(fract_names):
            mt_partial_volumes[partial_volume_name] = mt_fract_vec[..., i]
        return mt_partial_volumes

    def predict(self, acquisition_scheme=None, S0=None, mask=None):
        """
        simulates the dMRI signal of the fitted MultiCompartmentModel for the
        estimated model parameters. If no acquisition_scheme is given, then
        the same acquisition_scheme that was used for the fitting is used. If
        no S0 is given then it is assumed to be the estimated one. If no mask
        is given then all voxels are assumed to have been fitted.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        S0 : None or float,
            Signal intensity without diffusion sensitization. If None, uses
            estimated SO from fitting process. If float, uses that value.
        mask : (N-1)-dimensional integer/boolean array of size (N_x, N_y, ...),
            mask of voxels to simulate data at.

        Returns
        -------
        predicted_signal : array of size (Ndata, N_DWIS),
            predicted DWIs for the given model parameters and acquisition
            scheme.
        """
        if acquisition_scheme is None:
            acquisition_scheme = self.model.scheme
        self.model._check_model_params_with_acquisition_params(
            acquisition_scheme)

        dataset_shape = self.fitted_parameters_vector.shape[:-1]
        if S0 is None:
            S0 = self.S0
        elif isinstance(S0, float):
            S0 = np.ones(dataset_shape) * S0
        if mask is None:
            mask = self.mask

        N_samples = len(acquisition_scheme.shell_bvalues)

        predicted_signal = np.zeros(np.r_[dataset_shape, N_samples])
        mask_pos = np.where(mask)
        for pos in zip(*mask_pos):
            parameters = self.model.parameter_vector_to_parameters(
                self.fitted_parameters_vector[pos])
            predicted_signal[pos] = self.model(
                acquisition_scheme, **parameters) * S0[pos]
        return predicted_signal

    def R2_coefficient_of_determination(self, data):
        "Calculates the R-squared of the model fit."
        Nshells = len(self.model.scheme.shell_bvalues)
        data_ = np.zeros(np.r_[data.shape[:-1], Nshells])
        for pos in zip(*np.where(self.mask)):
            data_[pos] = estimate_spherical_mean_multi_shell(
                data[pos] / self.S0[pos], self.model.scheme)

        y_hat = self.predict(S0=1.)
        y_bar = np.mean(data_, axis=-1)
        SStot = np.sum((data_ - y_bar[..., None]) ** 2, axis=-1)
        SSres = np.sum((data_ - y_hat) ** 2, axis=-1)
        R2 = 1 - SSres / SStot
        R2[~self.mask] = 0
        return R2

    def mean_squared_error(self, data):
        "Calculates the mean squared error of the model fit."
        Nshells = len(self.model.scheme.shell_bvalues)
        data_ = np.zeros(np.r_[data.shape[:-1], Nshells])
        for pos in zip(*np.where(self.mask)):
            data_[pos] = estimate_spherical_mean_multi_shell(
                data[pos] / self.S0[pos], self.model.scheme)

        y_hat = self.predict(S0=1.)
        mse = np.mean((data_ - y_hat) ** 2, axis=-1)
        mse[~self.mask] = 0
        return mse

    def return_parametric_fod_model(
            self, distribution='watson', Ncompartments=1):
        """
        Retuns parametric FOD model using the rotational harmonics of the
        fitted spherical mean model as the convolution kernel. It can be called
        with any implemented parametric distribution (Watson/Bingham) and for
        any number of compartments.

        Internally, the input models to the spherical mean model are given to
        a spherically distributed model where the parameter links are replayed
        such that the distributed model has the same parameter constraints as
        the spherical mean model. This distributed model now represents one
        compartment of "bundle". This bundle representation is copied
        Ncompartment times and given as input to a MultiCompartmentModel, where
        now the non-linear are all linked such that each bundle has the same
        convolution kernel. Finally, the FittedSphericalMeanModel parameters
        are given as fixed parameters for the kernel (the kernel will not be
        fitted while the FOD's distribution parameters are being optimized).

        The function returns a MultiCompartmentModel instance that can be
        interacted with as usual to fit dMRI data.

        Parameters
        ----------
        distribution: string,
            Choice of parametric spherical distribution.
            Can be 'watson', or 'bingham'.
        Ncompartments: integer,
            Number of bundles that will be fitted. Must be larger than zero.

        Returns
        -------
        mc_bundles_model: Dmipy MultiCompartmentModel instance,
            MultiCompartmentModel instance that can be used to estimate
            parametric FODs using the fitted spherical mean model as a kernel.
        """
        from .modeling_framework import MultiCompartmentModel
        from ..distributions import distribute_models

        if not isinstance(Ncompartments, int) or Ncompartments < 1:
            msg = 'Ncompartments must be integer larger or equal to one.'
            raise ValueError(msg)

        if distribution == 'watson':
            bundle = distribute_models.SD1WatsonDistributed(
                self.model.models)
            basename = 'SD1WatsonDistributed_'
        elif distribution == 'bingham':
            bundle = distribute_models.SD2BinghamDistributed(
                self.model.models)
            basename = 'SD2BinghamDistributed_'
        else:
            msg = '{} is not a valid distribution choice'.format(
                distribute_models)
            raise ValueError(msg)

        for link in self.model.parameter_links:
            param_to_delete = self.model._inverted_parameter_map[link[0],
                                                                 link[1]]
            if isinstance(link[2], T1_tortuosity):
                bundle.parameter_links.append(
                    [link[0], link[1], link[2], link[3][:-1]])
            elif link[2] is fractional_parameter:
                new_parameter_name = param_to_delete + '_fraction'
                bundle.parameter_ranges.update({new_parameter_name: [0., 1.]})
                bundle.parameter_scales.update({new_parameter_name: 1.})
                bundle.parameter_cardinality.update({new_parameter_name: 1})
                bundle.parameter_types.update({new_parameter_name: 'normal'})

                bundle._parameter_map.update(
                    {new_parameter_name: (None, 'fraction')})
                bundle._inverted_parameter_map.update(
                    {(None, 'fraction'): new_parameter_name})

                # add parmeter link to fractional parameter
                param_larger_than = self.model._inverted_parameter_map[
                    link[3][1][0], link[3][1][1]]

                model, name = bundle._parameter_map[param_to_delete]
                bundle.parameter_links.append(
                    [model, name, fractional_parameter, [
                        bundle._parameter_map[new_parameter_name],
                        bundle._parameter_map[param_larger_than]]])
            else:
                bundle.parameter_links.append(link)
            del bundle.parameter_ranges[param_to_delete]
            del bundle.parameter_cardinality[param_to_delete]
            del bundle.parameter_scales[param_to_delete]
            del bundle.parameter_types[param_to_delete]

        bundles = [bundle.copy() for i in range(Ncompartments)]
        mc_bundles_model = MultiCompartmentModel(bundles)
        parameter_pairs = []
        for smt_par_name in self.model.parameter_names:
            parameters = []
            parameters.append(smt_par_name)
            for mc_par_name in mc_bundles_model.parameter_names:
                if (mc_par_name.startswith(basename) and
                        mc_par_name.endswith(smt_par_name)):
                    parameters.append(mc_par_name)
            if len(parameters) > 1:
                parameter_pairs.append(parameters)

        for parameters in parameter_pairs:
            for i in range(2, Ncompartments + 1):
                mc_bundles_model.set_equal_parameter(parameters[1],
                                                     parameters[i])

        for parameters in parameter_pairs:
            smt_parameter_name = parameters[0]
            mc_parameter_name = parameters[1]
            mc_bundles_model.set_fixed_parameter(
                mc_parameter_name,
                self.fitted_parameters[smt_parameter_name])
        return mc_bundles_model

    def return_spherical_harmonics_fod_model(self, sh_order=8):
        """
        Retuns spherical harmonics FOD model using the rotational harmonics of
        the fitted spherical mean model as the convolution kernel.

        Internally, the input models to the spherical mean model are given to
        a MultiCompartmentSphericalHarmonicsModel where the parameter links are
        replayed such that the new model has the same parameter constraints as
        the spherical mean model. The FittedSphericalMeanModel parameters
        are given as fixed parameters for the kernel (the kernel will not be
        fitted while the FOD's coefficients are being optimized).

        The function returns a MultiCompartmentSphericalHarmonicsModel instance
        that can be interacted with as usual to fit dMRI data.

        Parameters
        ----------
        sh_order: even, positive integer,
            Spherical harmonics order of the FODs.

        Returns
        -------
        mc_bundles_model: Dmipy MultiCompartmentModel instance,
            MultiCompartmentModel instance that can be used to estimate
            parametric FODs using the fitted spherical mean model as a kernel.
        """
        from .modeling_framework import MultiCompartmentSphericalHarmonicsModel

        if sh_order < 0 or sh_order % 2 != 0:
            msg = 'sh_order must be an even, positive integer.'
            raise ValueError(msg)

        sh_model = MultiCompartmentSphericalHarmonicsModel(
            self.model.models, sh_order=sh_order)

        for link in self.model.parameter_links:
            param_to_delete = self.model._inverted_parameter_map[link[0],
                                                                 link[1]]
            if isinstance(link[2], T1_tortuosity):
                sh_model.parameter_links.append(
                    [link[0], link[1], link[2], link[3][:-1]])
            elif link[2] is fractional_parameter:
                new_parameter_name = param_to_delete + '_fraction'
                sh_model.parameter_ranges.update(
                    {new_parameter_name: [0., 1.]})
                sh_model.parameter_scales.update({new_parameter_name: 1.})
                sh_model.parameter_cardinality.update({new_parameter_name: 1})
                sh_model.parameter_types.update({new_parameter_name: 'normal'})

                sh_model._parameter_map.update(
                    {new_parameter_name: (None, 'fraction')})
                sh_model._inverted_parameter_map.update(
                    {(None, 'fraction'): new_parameter_name})

                # add parmeter link to fractional parameter
                param_larger_than = self.model._inverted_parameter_map[
                    link[3][1][0], link[3][1][1]]

                model, name = sh_model._parameter_map[param_to_delete]
                sh_model.parameter_links.append(
                    [model, name, fractional_parameter, [
                        sh_model._parameter_map[new_parameter_name],
                        sh_model._parameter_map[param_larger_than]]])
            else:
                sh_model.parameter_links.append(link)
            del sh_model.parameter_ranges[param_to_delete]
            del sh_model.parameter_cardinality[param_to_delete]
            del sh_model.parameter_scales[param_to_delete]
            del sh_model.parameter_types[param_to_delete]
            del sh_model.parameter_optimization_flags[param_to_delete]

        for smt_par_name in self.model.parameter_names:
            sh_model.set_fixed_parameter(
                smt_par_name, self.fitted_parameters[smt_par_name])
        return sh_model


class FittedMultiCompartmentSphericalHarmonicsModel:
    """
    The FittedMultiCompartmentModel instance contains information about the
    original MultiCompartmentModel, the estimated S0 values, the fitting mask
    and the fitted model parameters.

    Parameters
    ----------
    model : MultiCompartmentModel instance,
        A dmipy MultiCompartmentModel.
    S0 : array of size (Ndata,) or (N_data, N_DWIs),
        Array containing the estimated S0 values of the data. If data is 4D,
        then S0 is 3D if there is only one TE, and the same 4D size of the data
        if there are multiple TEs.
    mask : array of size (N_data,),
        boolean mask of voxels that were fitted.
    fitted_parameters_vector : array of size (N_data, Nparameters),
        fitted model parameters array.
    """

    def __init__(self, model, S0, mask, fitted_parameters_vector):
        self.model = model
        self.S0 = S0
        self.mask = mask
        self.fitted_parameters_vector = fitted_parameters_vector
        self.S0_responses = model.S0_responses.copy()
        self.max_S0_response = model.max_S0_response
        self.fit_S0_response = model.fit_S0_response
        self._fit_parameters = {
            'fit_S0_response': self.fit_S0_response,
            'S0_responses': self.S0_responses}

    @property
    def fitted_parameters(self):
        "Returns the fitted parameters as a dictionary."
        return self.model.parameter_vector_to_parameters(
            self.fitted_parameters_vector)

    @property
    def fitted_and_linked_parameters(self):
        "Returns the fitted and linked parameters as a dictionary."
        fitted_parameters = self.model.parameter_vector_to_parameters(
            self.fitted_parameters_vector)
        return self.model.add_linked_parameters_to_parameters(
            fitted_parameters)

    def fod(self, vertices):
        """
        Returns the Fiber Orientation Distribution if it is available.

        Parameters
        ----------
        vertices : array of size (Nvertices, 3),
            Array of cartesian unit vectors at which to sample the FOD.
        visual_odi_lower_bound : float,
            gives a lower bound to the Orientation Distribution Index (ODI) of
            FODs of Watson and Bingham distributions. This can be useful to
            visualize FOD fields where some FODs are extremely sharp.

        Returns
        -------
        fods : array of size (Ndata, Nvertices),
            the FODs of the fitted model, scaled by volume fraction.
        """
        theta, phi = cart2mu(vertices).T
        sh_matrix = real_sym_sh_mrtrix(self.model.sh_order, theta, phi)[0]
        dataset_shape = self.fitted_parameters_vector.shape[:-1]
        N_samples = len(vertices)
        fods = np.zeros(np.r_[dataset_shape, N_samples])
        mask_pos = np.where(self.mask)
        for pos in zip(*mask_pos):
            parameters = self.model.parameter_vector_to_parameters(
                self.fitted_parameters_vector[pos])
            sh_coef = parameters['sh_coeff']
            fods[pos] = np.dot(sh_matrix, sh_coef)
        return fods

    def fod_sh(self):
        """
        Returns the spherical harmonics coefficients of the Fiber Orientation
        Distribution (FOD) if it is available. Uses are 724 spherical
        tessellation to do the spherical harmonics transform.

        Parameters
        ----------
        sh_order : integer,
            the maximum spherical harmonics order of the coefficient expansion.
        basis_type : string,
            type of spherical harmonics basis to use for the expansion, see
            sh_to_sf_matrix for more info.

        Returns
        -------
        fods_sh : array of size (Ndata, Ncoefficients),
            spherical harmonics coefficients of the FODs, scaled by volume
            fraction.
        """
        return self.fitted_parameters['sh_coeff']

    def peaks_directions(self, sphere, max_peaks=5,
                         relative_peak_threshold=0.5,
                         min_separation_angle=25):
        """
        Returns peak directions for estimated FODs. Uses dipy's peak_directions
        function to get the local maximum on a sphere's tesselation.

        Parameters
        ----------
        sphere : Sphere
            The Sphere providing discrete directions for evaluation.
        max_peaks : integer
            The maximum number of peaks that is returned per fod.
        relative_peak_threshold : float in [0., 1.]
            Only peaks greater than min + relative_peak_threshold * scale are
            kept, where min = max(0, odf.min()) and scale = odf.max() - min.
        min_separation_angle : float in [0, 90]
            The minimum distance between directions. If two peaks are too close
            only the larger of the two is returned.

        Returns
        -------
        directions : (Ndata, Npeaks, 3,) ndarray
            N vertices for sphere, one for each peak
        values : (Ndata, Npeaks,) ndarray
            peak values
        indices : (Ndata, Npeaks,) ndarray
            peak indices of the directions on the sphere
        """
        peaks_shape = np.r_[self.mask.shape, max_peaks, 3]
        peaks = np.zeros(peaks_shape)
        values = np.zeros(peaks_shape[:-1])
        indices = np.zeros(peaks_shape[:-1])

        fods = self.fod(sphere.vertices)
        mask_pos = np.where(self.mask)
        for pos in zip(*mask_pos):
            fod = fods[pos]
            peaks_, values_, indices_ = peak_directions(
                fod, sphere, relative_peak_threshold, min_separation_angle)
            # if less peaks than max_peaks are found, only take those.
            Npeaks = np.min([len(indices_), max_peaks])
            peaks[pos][:Npeaks] = peaks_[:Npeaks]
            values[pos][:Npeaks] = values_[:Npeaks]
            indices[pos][:Npeaks] = indices_[:Npeaks]
        return peaks, values, indices

    def anisotropy_index(self):
        """
        Estimates anisotropy index of spherical harmonics [1]_.

        References
        ----------
        .. [1] Jespersen, Sune N., et al. "Modeling dendrite density from
            magnetic resonance diffusion measurements." Neuroimage 34.4 (2007):
            1473-1486.

        """
        sh_coef = self.fitted_parameters['sh_coeff']
        sh_0 = sh_coef[..., 0] ** 2
        sh_sum_squared = np.sum(sh_coef ** 2, axis=-1)
        AI = np.zeros_like(sh_0)
        AI[self.mask] = np.sqrt(1 - sh_0[self.mask] /
                                sh_sum_squared[self.mask])
        return AI

    def norm_of_laplacian_fod(self):
        """
        Estimates the squared norm of the Laplacian of the FOD. Similar to
        the anisotropy index, a higher norm means there are larger higher-order
        coefficients in the FOD spherical harmonics expansion. This indicates
        the FOD is more anisotropic overall. This kind of measure was explored
        in e.g. [1]_.

        References
        ----------
        .. [1] Descoteaux, Maxime, et al. "Regularized, fast, and robust
            analytical Q-ball imaging." Magnetic Resonance in Medicine: An
            Official Journal of the International Society for Magnetic
            Resonance in Medicine 58.3 (2007): 497-510.
        """
        sh_coef = self.fitted_parameters['sh_coeff']
        sh_l = sph_harm_ind_list(self.model.sh_order)[1]
        lb_weights = sh_l ** 2 * (sh_l + 1) ** 2  # laplace-beltrami
        norm_laplacian = np.sum(sh_coef ** 2 * lb_weights, axis=-1)
        return norm_laplacian

    def predict(self, acquisition_scheme=None, S0=None, mask=None):
        """
        simulates the dMRI signal of the fitted MultiCompartmentModel for the
        estimated model parameters. If no acquisition_scheme is given, then
        the same acquisition_scheme that was used for the fitting is used. If
        no S0 is given then it is assumed to be the estimated one. If no mask
        is given then all voxels are assumed to have been fitted.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        S0 : None or float,
            Signal intensity without diffusion sensitization. If None, uses
            estimated SO from fitting process. If float, uses that value.
        mask : (N-1)-dimensional integer/boolean array of size (N_x, N_y, ...),
            mask of voxels to simulate data at.

        Returns
        -------
        predicted_signal : array of size (Ndata, N_DWIS),
            predicted DWIs for the given model parameters and acquisition
            scheme.
        """
        if acquisition_scheme is None:
            acquisition_scheme = self.model.scheme
        self.model._check_model_params_with_acquisition_params(
            acquisition_scheme)

        dataset_shape = self.fitted_parameters_vector.shape[:-1]
        if S0 is None:
            if self.fit_S0_response:
                S0 = np.ones(dataset_shape) * self.max_S0_response
            else:
                S0 = self.S0
        elif isinstance(S0, float):
            if self.fit_S0_response:
                S0 = S0 * self.max_S0_response / self.S0
            else:
                S0 = np.ones(dataset_shape) * S0
        if mask is None:
            mask = self.mask

        N_samples = len(acquisition_scheme.bvalues)

        predicted_signal = np.zeros(np.r_[dataset_shape, N_samples])
        mask_pos = np.where(mask)
        for pos in zip(*mask_pos):
            parameters = self.model.parameter_vector_to_parameters(
                self.fitted_parameters_vector[pos])
            predicted_signal[pos] = (
                self.model(acquisition_scheme,
                           **dict(parameters,
                                  **self._fit_parameters)) * S0[pos])
        return predicted_signal

    def R2_coefficient_of_determination(self, data):
        "Calculates the R-squared of the model fit."
        if self.model.scheme.TE is None:
            data_ = data / self.S0[..., None]
        else:
            data_ = data / self.S0

        y_hat = self.predict(S0=1.)
        y_bar = np.mean(data_, axis=-1)
        SStot = np.sum((data_ - y_bar[..., None]) ** 2, axis=-1)
        SSres = np.sum((data_ - y_hat) ** 2, axis=-1)
        R2 = 1 - SSres / SStot
        R2[~self.mask] = 0
        return R2

    def mean_squared_error(self, data):
        "Calculates the mean squared error of the model fit."
        if self.model.scheme.TE is None:
            data_ = data / self.S0[..., None]
        else:
            data_ = data / self.S0

        y_hat = self.predict(S0=1.)
        mse = np.mean((data_ - y_hat) ** 2, axis=-1)
        mse[~self.mask] = 0
        return mse



import copy
from collections import OrderedDict
from itertools import chain

import numpy as np

from . import distributions
from ..core.signal_model_properties import AnisotropicSignalModelProperties
from ..utils.spherical_convolution import sh_convolution
from ..utils.utils import T1_tortuosity, parameter_equality

__all__ = [
    'DistributedModel',
    'SD1WatsonDistributed',
    'SD2BinghamDistributed',
    'DD1GammaDistributed',
    'ReturnFixedValue'
]


class DistributedModel:
    "Contains various properties of distributed models."
    _spherical_mean = False

    def _check_for_double_model_class_instances(self):
        "Checks if there are no models with the same class instantiation."
        if len(self.models) != len(set(self.models)):
            msg = "Each model in the multi-compartment model must be "
            msg += "instantiated separately. For example, to make a model "
            msg += "with two sticks, the models must be given as "
            msg += "models = [stick1, stick2], not as "
            msg += "models = [stick1, stick1]."
            raise ValueError(msg)

    def _check_for_dispersable_models(self):
        """
        Checks if a model can be dispersed, i.e. has a rotational harmonics
        representation.
        """
        for i, model in enumerate(self.models):
            try:
                callable(model.rotational_harmonics_representation)
            except AttributeError:
                msg = "Cannot disperse models input {}. ".format(i)
                msg += "It has no rotational_harmonics_representation."
                raise AttributeError(msg)

    def _check_for_distributable_models(self):
        "Checks if the to-be distributed parameter is in the input models."
        for i, model in enumerate(self.models):
            if self.target_parameter not in model.parameter_ranges:
                msg = "Cannot distribute models input {}. ".format(i)
                msg += "It has no {} as parameter.".format(
                    self.target_parameter)
                raise AttributeError(msg)

    def _check_for_same_parameter_type(self):
        """
        For gamma distribution, checks if sphere/cylinder models are not
        mixed.
        """
        parameter_types = [model.parameter_types[self.target_parameter]
                           for model in self.models]
        if len(np.unique(parameter_types)) > 1:
            msg = "Cannot mix models with different parameter types. "
            msg += "Current input parameter types are {}.".format(
                parameter_types)
            raise AttributeError(msg)

    def _prepare_parameters(self, models_and_distribution):
        """
        Prepares the DistributedModel parameter properties for both the
        distribution and the distributed models.
        """
        self.model_names = []
        model_counts = {}

        for model in models_and_distribution:
            if model.__class__ not in model_counts:
                model_counts[model.__class__] = 1
            else:
                model_counts[model.__class__] += 1

            self.model_names.append(
                '{}_{:d}_'.format(
                    model.__class__.__name__,
                    model_counts[model.__class__]
                )
            )

        self.parameter_ranges = OrderedDict({
            model_name + k: v
            for model, model_name in zip(models_and_distribution,
                                         self.model_names)
            for k, v in model.parameter_ranges.items()
        })

        self.parameter_scales = OrderedDict({
            model_name + k: v
            for model, model_name in zip(models_and_distribution,
                                         self.model_names)
            for k, v in model.parameter_scales.items()
        })

        self.parameter_types = OrderedDict({
            model_name + k: v
            for model, model_name in zip(models_and_distribution,
                                         self.model_names)
            for k, v in model.parameter_types.items()
        })

        self._parameter_map = {
            model_name + k: (model, k)
            for model, model_name in zip(models_and_distribution,
                                         self.model_names)
            for k in model.parameter_ranges
        }

        self._inverted_parameter_map = {
            v: k for k, v in self._parameter_map.items()
        }

        self.parameter_cardinality = OrderedDict([
            (k, len(np.atleast_2d(self.parameter_ranges[k])))
            for k in self.parameter_ranges
        ])

    def _delete_orientation_from_parameters(self):
        "Removes orientation parameters from input models."
        orientation_counter = 0
        for model in self.models:
            for param_name, param_type in model.parameter_types.items():
                if param_type == 'orientation':
                    appended_param_name = self._inverted_parameter_map[
                        model, param_name]
                    del self.parameter_ranges[appended_param_name]
                    del self.parameter_scales[appended_param_name]
                    del self.parameter_cardinality[appended_param_name]
                    del self.parameter_types[appended_param_name]
                    orientation_counter += 1
        if orientation_counter == 0:
            msg = 'At least one input model must have an orientation '
            msg += 'parameter.'
            raise ValueError(msg)

    def _delete_target_parameter_from_parameters(self):
        "Removes the to-be-distributed parameter from the parameter list."
        for model in self.models:
            parameter_name = self._inverted_parameter_map[
                (model, self.target_parameter)
            ]
            del self.parameter_ranges[parameter_name]
            del self.parameter_scales[parameter_name]
            del self.parameter_cardinality[parameter_name]
            del self.parameter_types[parameter_name]

    def _prepare_partial_volumes(self):
        "Prepares the partial volumes for the DistributedModel."
        if len(self.models) > 1:
            self.partial_volume_names = [
                'partial_volume_{:d}'.format(i)
                for i in range(len(self.models) - 1)
            ]

            for i, partial_volume_name in enumerate(self.partial_volume_names):
                self.parameter_ranges[partial_volume_name] = (0.01, .99)
                self.parameter_scales[partial_volume_name] = 1.
                self._parameter_map[partial_volume_name] = (
                    None, partial_volume_name
                )
                self.parameter_types[partial_volume_name] = 'normal'
                self._inverted_parameter_map[(None, partial_volume_name)] = \
                    partial_volume_name
                self.parameter_cardinality[partial_volume_name] = 1

    def _prepare_parameter_links(self):
        "Prepares the parameter links, if they are given."
        for i, parameter_function in enumerate(self.parameter_links):
            parameter_model, parameter_name, parameter_function, arguments = \
                parameter_function

            if (
                    (parameter_model, parameter_name)
                    not in self._inverted_parameter_map
            ):
                raise ValueError(
                    "Parameter function {} doesn't exist".format(i)
                )

            parameter_name = self._inverted_parameter_map[
                (parameter_model, parameter_name)
            ]

            del self.parameter_ranges[parameter_name]
            del self.parameter_scales[parameter_name]
            del self.parameter_cardinality[parameter_name]
            del self.parameter_types[parameter_name]

    def add_linked_parameters_to_parameters(self, parameters):
        """
        Adds the linked parameters to the optimized parameters.

        Parameters
        ----------
        parameters: dictionary of model parameters,
            contains the optimized parameters.

        Returns
        -------
        parameters: dictionary of model parameters,
            contains the optimzed and linked parameters.
        """
        if len(self.parameter_links) == 0:
            return parameters
        parameters = parameters.copy()
        for parameter in self.parameter_links[::-1]:
            parameter_model, parameter_name, parameter_function, arguments = \
                parameter
            parameter_name = self._inverted_parameter_map[
                (parameter_model, parameter_name)
            ]

            if len(arguments) > 0:
                argument_values = []
                for argument in arguments:
                    argument_name = self._inverted_parameter_map[argument]
                    argument_values.append(parameters.get(
                        argument_name
                    ))

                parameters[parameter_name] = parameter_function(
                    *argument_values
                )
            else:
                parameters[parameter_name] = parameter_function()
        return parameters

    def set_fixed_parameter(self, parameter_name, value):
        """
        Allows the user to fix an optimization parameter to a static value.
        The fixed parameter will be removed from the optimized parameters and
        added as a linked parameter.

        Parameters
        ----------
        parameter_name: string
            name of the to-be-fixed parameters, see self.parameter_names.
        value: float or list of corresponding parameter_cardinality.
            the value to fix the parameter at in SI units.
        """
        if parameter_name in self.parameter_ranges.keys():
            card = self.parameter_cardinality[parameter_name]
            if card == 1:
                if isinstance(value, int):
                    value = float(value)
                if not isinstance(value, float):
                    msg = '{} can only be fixed to a float value.'.format(
                        parameter_name)
                    raise ValueError(msg)
            elif card == 2:
                value = np.array(value, dtype=float)
                if value.shape != (2,):
                    msg = '{} can only be fixed '.format(parameter_name)
                    msg += 'to an array or list of length 2.'
                    raise ValueError(msg)
            model, name = self._parameter_map[parameter_name]
            parameter_link = (model, name, ReturnFixedValue(value), [])
            self.parameter_links.append(parameter_link)
            del self.parameter_ranges[parameter_name]
            del self.parameter_scales[parameter_name]
            del self.parameter_cardinality[parameter_name]
            del self.parameter_types[parameter_name]
        else:
            msg = '{} does not exist or has already been fixed.'.format(
                parameter_name)
            raise ValueError(msg)

    def set_tortuous_parameter(self, lambda_perp,
                               lambda_par,
                               volume_fraction_intra):
        """
        Allows the user to set a tortuosity constraint on the perpendicular
        diffusivity of the extra-axonal compartment, which depends on the
        intra-axonal volume fraction and parallel diffusivity.

        The perpendicular diffusivity parameter will be removed from the
        optimized parameters and added as a linked parameter.

        Parameters
        ----------
        lambda_perp: string
            name of the perpendicular diffusivity parameter, see
            self.parameter_names.
        lambda_par: string
            name of the parallel diffusivity parameter, see
            self.parameter_names.
        volume_fraction_intra: string
            name of the intra-axonal volume fraction parameter, see
            self.parameter_names.
        """
        params = [lambda_perp, lambda_par, volume_fraction_intra]
        for param in params:
            try:
                self.parameter_ranges[param]
            except KeyError:
                print("{} does not exist or has already been fixed.").format(
                    param)
                return None

        tortuosity = T1_tortuosity()
        model, name = self._parameter_map[lambda_perp]
        self.parameter_links.append([model, name, tortuosity, [
            self._parameter_map[lambda_par],
            self._parameter_map[volume_fraction_intra]]])
        del self.parameter_ranges[lambda_perp]
        del self.parameter_scales[lambda_perp]
        del self.parameter_cardinality[lambda_perp]
        del self.parameter_types[lambda_perp]

    def set_equal_parameter(self, parameter_name_in, parameter_name_out):
        """
        Allows the user to set two parameters equal to each other. This is used
        for example in the NODDI model to set the parallel diffusivities of the
        Stick and Zeppelin compartment to the same value.

        The second input parameter will be removed from the optimized
        parameters and added as a linked parameter.

        Parameters
        ----------
        parameter_name_in: string
            the first parameter name, see self.parameter_names.
        parameter_name_out: string,
            the second parameter name, see self.parameter_names. This is the
            parameter that will be removed form the optimzed parameters.
        """
        params = [parameter_name_in, parameter_name_out]
        for param in params:
            try:
                self.parameter_ranges[param]
            except KeyError:
                print("{} does not exist or has already been fixed.".format(
                    param))
                return None
        model, name = self._parameter_map[parameter_name_out]
        self.parameter_links.append([model, name, parameter_equality, [
            self._parameter_map[parameter_name_in]]])
        del self.parameter_ranges[parameter_name_out]
        del self.parameter_scales[parameter_name_out]
        del self.parameter_cardinality[parameter_name_out]
        del self.parameter_types[parameter_name_out]

    def copy(self):
        """
        Retuns a different instantiation of the DistributedModel with the same
        configuration, which can be used together with the original in a
        MultiCompartmentModel. For example, to do NODDI with multiple
        orientations.
        """
        return copy.copy(self)

    def _set_required_acquisition_parameters(self):
        self._required_acquisition_parameters = []
        for model in self.models:
            self._required_acquisition_parameters += (
                model._required_acquisition_parameters)
        self._required_acquisition_parameters = np.unique(
            self._required_acquisition_parameters).tolist()

    @property
    def parameter_names(self):
        "Retuns the DistributedModel parameter names."
        return self.parameter_ranges.keys()

    def __call__(self, acquisition_scheme, **kwargs):
        """
        The DistributedModel function call. If the model is spherical
        distribution it will return a spherical harmonics-convolved model.
        If it is a spatial distribution then it returns an integrated model.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.
        """
        if hasattr(self, 'distribution'):
            if self._model_type == 'SphericalDistributedModel':
                return self.sh_convolved_model(acquisition_scheme, **kwargs)
            elif self._model_type == 'SpatialDistributedModel':
                return self.integrated_model(acquisition_scheme, **kwargs)
            else:
                raise RuntimeError(
                    'Cannot generate signal for unknown distribution type '
                    '{}'.format(self._model_type))
        else:
            return self.bundle_model(acquisition_scheme, **kwargs)

    def bundle_model(self, acquisition_scheme, **kwargs):
        """
        Simple bundle model that does not apply any distribution. It can be
        considered a sub-multi-compartment model that allows for tortuosity
        constraints.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.
        """
        kwargs = self.add_linked_parameters_to_parameters(
            kwargs
        )

        if len(self.models) > 1:
            partial_volumes = [
                kwargs[p] for p in self.partial_volume_names
            ]
        else:
            partial_volumes = []

        remaining_volume_fraction = 1.
        E = 0.
        for model_name, model, partial_volume in zip(
                self.model_names, self.models,
                chain(partial_volumes, [None])
        ):
            parameters = {}
            for parameter in model.parameter_ranges:
                parameter_name = self._inverted_parameter_map[
                    (model, parameter)
                ]
                parameters[parameter] = kwargs.get(
                    parameter_name
                )

            if partial_volume is not None:
                volume_fraction = remaining_volume_fraction * partial_volume
                remaining_volume_fraction = (
                    remaining_volume_fraction - volume_fraction)
            else:
                volume_fraction = remaining_volume_fraction
            E += volume_fraction * model(acquisition_scheme, **parameters)
        return E

    def sh_convolved_model(self, acquisition_scheme, **kwargs):
        """
        The spherical harmonics convolved function call for spherically
        distributions like Watson and Bingham.

        First, the linked parameters are added to the optimized parameters, and
        the spherical harmonics of the spherical distribution are recovered.
        The volume fractions are also converted from nested to regular ones.

        Then, for every model in the DistributedModel, and for every
        acquisition shell, the rotational harmonics of the model are recovered
        and convolved with the distribution spherical harmonics. The resulting
        values are multiplied with the volume fractions and finally the
        dispersed signal attenuation is returned.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.
        """
        kwargs = self.add_linked_parameters_to_parameters(
            kwargs
        )

        distribution_parameters = {}
        for parameter in self.distribution.parameter_ranges:
            parameter_name = self._inverted_parameter_map[
                (self.distribution, parameter)
            ]
            distribution_parameters[parameter] = kwargs.get(
                parameter_name)
        sh_distribution = self.distribution.spherical_harmonics_representation(
            **distribution_parameters)

        if len(self.models) > 1:
            partial_volumes = [
                kwargs[p] for p in self.partial_volume_names
            ]
        else:
            partial_volumes = []

        remaining_volume_fraction = 1.
        rh_models = 0.
        for model_name, model, partial_volume in zip(
                self.model_names, self.models,
                chain(partial_volumes, [None])
        ):
            parameters = {}
            for parameter in model.parameter_ranges:
                parameter_name = self._inverted_parameter_map[
                    (model, parameter)
                ]
                parameters[parameter] = kwargs.get(
                    parameter_name
                )

            if partial_volume is not None:
                volume_fraction = remaining_volume_fraction * partial_volume
                remaining_volume_fraction = (
                    remaining_volume_fraction - volume_fraction)
            else:
                volume_fraction = remaining_volume_fraction

            rh_model = model.rotational_harmonics_representation(
                acquisition_scheme, **parameters)
            rh_models = rh_models + volume_fraction * rh_model

        E = np.ones(acquisition_scheme.number_of_measurements)
        for i, shell_index in enumerate(acquisition_scheme.unique_dwi_indices):
            shell_mask = acquisition_scheme.shell_indices == shell_index
            sh_mat = acquisition_scheme.shell_sh_matrices[shell_index]
            sh_order = int(acquisition_scheme.shell_sh_orders[shell_index])
            E_dispersed_sh = sh_convolution(
                sh_distribution, rh_models[i, :sh_order // 2 + 1])
            E[shell_mask] = np.dot(sh_mat[:, :len(E_dispersed_sh)],
                                   E_dispersed_sh)
        return E

    def integrated_model(self, acquisition_scheme, **kwargs):
        """
        The spatially integrated function call for spatial distributions like
        Gamma. Currently, the model assumed we are distributing diameters.

        First, the linked parameters are added to the optimized parameters, and
        the probability that a water particle exists inside a cylinder of a
        certain size in the distribution is sampled for a range of diameters.
        The volume fractions are also converted from nested to regular ones
        (although typically not more than 1 model is used for a Gamma
        distribution).

        Then, for every model in the DistributedModel, the signal attenuations
        of the model are are recovered for the sampled diameters and multiplied
        and integrated over their probabilities. The resulting
        values are multiplied with the volume fractions and finally the
        integrated signal attenuation is returned.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.
        """
        values = 0.
        kwargs = self.add_linked_parameters_to_parameters(
            kwargs
        )

        distribution_parameters = {}
        for parameter in self.distribution.parameter_ranges:
            parameter_name = self._inverted_parameter_map[
                (self.distribution, parameter)
            ]
            distribution_parameters[parameter] = kwargs.get(
                parameter_name)
        radii, P_radii = self.distribution(**distribution_parameters)

        if len(self.models) > 1:
            partial_volumes = [
                kwargs[p] for p in self.partial_volume_names
            ]
        else:
            partial_volumes = []
        remaining_volume_fraction = 1.
        for model_name, model, partial_volume in zip(
                self.model_names, self.models,
                chain(partial_volumes, [None])
        ):
            parameters = {}
            for parameter in model.parameter_ranges:
                parameter_name = self._inverted_parameter_map[
                    (model, parameter)
                ]
                parameters[parameter] = kwargs.get(
                    parameter_name
                )
            E = np.empty(
                (len(radii),
                 acquisition_scheme.number_of_measurements))
            for i, radius in enumerate(radii):
                parameters[self.target_parameter] = radius * 2
                E[i] = P_radii[i] * model(acquisition_scheme, **parameters)
            E = np.trapz(E, x=radii, axis=0)

            if partial_volume is not None:
                volume_fraction = remaining_volume_fraction * partial_volume
                remaining_volume_fraction = (
                    remaining_volume_fraction - volume_fraction)
            else:
                volume_fraction = remaining_volume_fraction

            values = values + volume_fraction * E
        return values

    def fod(self, vertices, **kwargs):
        """
        Returns the Fiber Orientation Distribution (FOD) of a dispersed model.

        Parameters
        ----------
        vertices: array of size (N_samples, 3)
            cartesian unit vectors at which to sample the FOD
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.
        """
        distribution_parameters = {}
        for parameter in self.distribution.parameter_ranges:
            parameter_name = self._inverted_parameter_map[
                (self.distribution, parameter)
            ]
            distribution_parameters[parameter] = kwargs.get(
                parameter_name)
        return self.distribution(vertices, **distribution_parameters)


class BundleModel(DistributedModel, AnisotropicSignalModelProperties):
    """
    The DistributedModel instantiation for a simple, non-distributed bundle
    model. This allows to join models in a secondary layer that allows for
    tortuosity constraints.

    Parameters
    ----------
    models: list of length 1 or more,
        list of models to be Watson-dispersed.
    parameters_links: list of length 1 or more,
        deprecated for testing use only.
    """
    _model_type = 'BundleModel'

    def __init__(self, models, parameter_links=None):
        self.models = models
        self._set_required_acquisition_parameters()
        self._check_for_double_model_class_instances()

        self.parameter_links = parameter_links
        if parameter_links is None:
            self.parameter_links = []

        _models_and_distribution = list(self.models)
        self._prepare_parameters(_models_and_distribution)
        self._prepare_partial_volumes()
        self._prepare_parameter_links()
        self.mu_params = []
        for param in self.parameter_names:
            if param.endswith('mu'):
                self.mu_params.append(param)

    def rotational_harmonics_representation(
            self, acquisition_scheme, **kwargs):
        r""" The rotational harmonics of the model, such that Y_lm = Yl0.
        Axis aligned with z-axis to be used as kernel for spherical
        convolution. Returns an array with rotational harmonics for each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        rh_array : array, shape(Nshells, N_rh_coef),
            Rotational harmonics coefficients for each shell.
        """
        rh_scheme = acquisition_scheme.rotational_harmonics_scheme
        rh_scheme.rotational_harmonics_scheme = rh_scheme
        for param in self.mu_params:
            kwargs.update({param: [0., 0.]})
        E_kernel_sf = self(rh_scheme, **kwargs)
        E_reshaped = E_kernel_sf.reshape([-1, rh_scheme.Nsamples])
        max_sh_order = max(rh_scheme.shell_sh_orders.values())
        rh_array = np.zeros((len(E_reshaped), max_sh_order // 2 + 1))

        for i, (shell_index, sh_order) in enumerate(
                rh_scheme.shell_sh_orders.items()):
            rh_array[i, :sh_order // 2 + 1] = (
                np.dot(
                    rh_scheme.inverse_rh_matrix[sh_order],
                    E_reshaped[i])
            )
        return rh_array


class SD1WatsonDistributed(DistributedModel, AnisotropicSignalModelProperties):
    """
    The DistributedModel instantiation for a Watson-dispersed model. Multiple
    models can be dispersed at the same time (like a Stick and Zeppelin for
    NODDI for example). The parameter 'mu' of the models will be removed and
    replaced by the 'mu' and 'odi' of the Watson distribution.

    After instantiation the WatsonDistributed model can be treated exactly the
    same as a CompartmentModel as an input for a MultiCompartmentModel.

    Parameters
    ----------
    models: list of length 1 or more,
        list of models to be Watson-dispersed.
    parameters_links: list of length 1 or more,
        deprecated for testing use only.
    """
    _model_type = 'SphericalDistributedModel'

    def __init__(self, models, parameter_links=None):
        self.models = models
        self._set_required_acquisition_parameters()
        self._check_for_double_model_class_instances()
        self._check_for_dispersable_models()

        self.parameter_links = parameter_links
        if parameter_links is None:
            self.parameter_links = []
        self.distribution = distributions.SD1Watson()

        _models_and_distribution = list(self.models)
        _models_and_distribution.append(self.distribution)
        self._prepare_parameters(_models_and_distribution)
        self._delete_orientation_from_parameters()
        self._prepare_partial_volumes()
        self._prepare_parameter_links()
        for param in self.parameter_names:
            if param.endswith('mu'):
                self.mu_param = param

    def rotational_harmonics_representation(
            self, acquisition_scheme, **kwargs):
        r""" The rotational harmonics of the model, such that Y_lm = Yl0.
        Axis aligned with z-axis to be used as kernel for spherical
        convolution. Returns an array with rotational harmonics for each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        rh_array : array, shape(Nshells, N_rh_coef),
            Rotational harmonics coefficients for each shell.
        """
        rh_scheme = acquisition_scheme.rotational_harmonics_scheme
        rh_scheme.rotational_harmonics_scheme = rh_scheme
        kwargs.update({self.mu_param: [0., 0.]})
        E_kernel_sf = self(rh_scheme, **kwargs)
        E_reshaped = E_kernel_sf.reshape([-1, rh_scheme.Nsamples])
        max_sh_order = max(rh_scheme.shell_sh_orders.values())
        rh_array = np.zeros((len(E_reshaped), max_sh_order // 2 + 1))

        for i, (shell_index, sh_order) in enumerate(
                rh_scheme.shell_sh_orders.items()):
            rh_array[i, :sh_order // 2 + 1] = (
                np.dot(
                    rh_scheme.inverse_rh_matrix[sh_order],
                    E_reshaped[i])
            )
        return rh_array


class SD2BinghamDistributed(DistributedModel,
                            AnisotropicSignalModelProperties):
    """
    The DistributedModel instantiation for a Bingham-dispersed model. Multiple
    models can be dispersed at the same time (like a Stick and Zeppelin for
    NODDI for example). The parameter 'mu' of the models will be removed and
    replaced by the 'mu', 'odi', 'beta_fraction' and 'psi' of the Bingham
    distribution.

    After instantiation the BinghamDistributed model can be treated exactly the
    same as a CompartmentModel as an input for a MultiCompartmentModel.

    Parameters
    ----------
    models: list of length 1 or more,
        list of models to be Bingham-dispersed.
    parameters_links: list of length 1 or more,
        deprecated for testing use only.
    """
    _model_type = 'SphericalDistributedModel'

    def __init__(self, models, parameter_links=None):
        self.models = models
        self._set_required_acquisition_parameters()
        self._check_for_double_model_class_instances()
        self._check_for_dispersable_models()

        self.parameter_links = parameter_links
        if parameter_links is None:
            self.parameter_links = []
        self.distribution = distributions.SD2Bingham()

        _models_and_distribution = list(self.models)
        _models_and_distribution.append(self.distribution)
        self._prepare_parameters(_models_and_distribution)
        self._delete_orientation_from_parameters()
        self._prepare_partial_volumes()
        self._prepare_parameter_links()
        for param in self.parameter_names:
            if param.endswith('mu'):
                self.mu_param = param


class SD3SphericalHarmonicsDistributed(DistributedModel,
                                       AnisotropicSignalModelProperties):
    """
    The DistributedModel instantiation of the SphericalHarmonicsDistributed
    model. Multiple models can be dispersed at the same time.

    Its use is intended for estimating the convolution kernel given some
    distribution in the regular MC-model. To fit spherical harmonics of a
    distributed model one must use the MC-SH approach.

    To estimate the kernel, however, The SH parameters must be fixed before
    fitting the model

    Parameters
    ----------
    models: list of length 1 or more,
        list of models to be SH-dispersed.
    sh_order: int,
        maximum spherical harmonics order to use.
    parameters_links: list of length 1 or more,
        deprecated for testing use only.
    """

    _model_type = 'SphericalDistributedModel'

    def __init__(self, models, sh_order, parameter_links=None):
        self.models = models
        self._set_required_acquisition_parameters()
        self._check_for_double_model_class_instances()
        self._check_for_dispersable_models()
        self.parameter_links = parameter_links
        if parameter_links is None:
            self.parameter_links = []
        self.distribution = distributions.SD3SphericalHarmonics(sh_order)

        _models_and_distribution = list(self.models)
        _models_and_distribution.append(self.distribution)
        self._prepare_parameters(_models_and_distribution)
        self._delete_orientation_from_parameters()
        self._prepare_partial_volumes()
        self._prepare_parameter_links()
        for param in self.parameter_names:
            if param.endswith('mu'):
                self.mu_param = param


class DD1GammaDistributed(DistributedModel, AnisotropicSignalModelProperties):
    """
    The DistributedModel instantiation for a Gamma-distributed model for
    cylinder or sphere models. Multiple models can be distributed at the same
    time (but such multi-cylinder-distribution models have never been used as
    far as we know). The parameter 'diameter' of the models will be removed and
    replaced by the 'alpha' and 'beta', of the Gamma distribution.

    After instantiation the GammaDistributed model can be treated exactly the
    same as a CompartmentModel as an input for a MultiCompartmentModel.

    Parameters
    ----------
    models: list of length 1 or more,
        list of models to be Watson-dispersed.
    parameters_links: list of length 1 or more,
        deprecated for testing use only.
    """
    _model_type = 'SpatialDistributedModel'

    def __init__(self, models, parameter_links=None,
                 target_parameter='diameter'):
        self.models = models
        self._set_required_acquisition_parameters()
        self.target_parameter = target_parameter
        self._check_for_double_model_class_instances()
        self._check_for_distributable_models()
        self._check_for_same_parameter_type()

        self.parameter_links = parameter_links
        if parameter_links is None:
            self.parameter_links = []

        self.normalization = models[0].parameter_types[target_parameter]
        self.distribution = distributions.DD1Gamma(
            normalization=self.normalization)

        _models_and_distribution = list(self.models)
        _models_and_distribution.append(self.distribution)
        self._prepare_parameters(_models_and_distribution)
        self._delete_target_parameter_from_parameters()
        self._prepare_partial_volumes()
        self._prepare_parameter_links()
        for param in self.parameter_names:
            if param.endswith('mu'):
                self.mu_param = param

    def rotational_harmonics_representation(
            self, acquisition_scheme, **kwargs):
        r""" The rotational harmonics of the model, such that Y_lm = Yl0.
        Axis aligned with z-axis to be used as kernel for spherical
        convolution. Returns an array with rotational harmonics for each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        rh_array : array, shape(Nshells, N_rh_coef),
            Rotational harmonics coefficients for each shell.
        """
        rh_scheme = acquisition_scheme.rotational_harmonics_scheme
        kwargs.update({self.mu_param: [0., 0.]})
        E_kernel_sf = self(rh_scheme, **kwargs)
        E_reshaped = E_kernel_sf.reshape([-1, rh_scheme.Nsamples])
        max_sh_order = max(rh_scheme.shell_sh_orders.values())
        rh_array = np.zeros((len(E_reshaped), max_sh_order // 2 + 1))

        for i, (shell_index, sh_order) in enumerate(
                rh_scheme.shell_sh_orders.items()):
            rh_array[i, :sh_order // 2 + 1] = (
                np.dot(
                    rh_scheme.inverse_rh_matrix[sh_order],
                    E_reshaped[i])
            )
        return rh_array

    def set_diameter_constrained_parameter_beta(
            self, diameter_min, diameter_max):
        # append new parameter to parameters
        for parameter in self.parameter_names:
            if parameter.endswith('beta'):
                beta_param = parameter
            if parameter.endswith('alpha'):
                alpha_param = parameter
        new_parameter_name = beta_param + '_fraction'
        parameters = [alpha_param, new_parameter_name]
        self.parameter_ranges.update({new_parameter_name: [0., 1.]})
        self.parameter_scales.update({new_parameter_name: 1.})
        self.parameter_cardinality.update({new_parameter_name: 1})
        self.parameter_types.update({new_parameter_name: 'normal'})

        self._parameter_map.update({new_parameter_name: (None, 'fraction')})
        self._inverted_parameter_map.update(
            {(None, 'fraction'): new_parameter_name})

        # add parmeter link to fractional parameter
        opt_function = ReturnConstrainedBeta(diameter_min, diameter_max)
        model, name = self._parameter_map[beta_param]
        self.parameter_links.append([model, name, opt_function,
                                     [self._parameter_map[param]
                                      for param in parameters]])
        del self.parameter_ranges[beta_param]
        del self.parameter_scales[beta_param]
        del self.parameter_cardinality[beta_param]
        del self.parameter_types[beta_param]


class ReturnConstrainedBeta:
    "Optimization parameter class to constrain gamma distribution mean."

    def __init__(self, diameter_min, diameter_max):
        self.diameter_min = diameter_min
        self.diameter_max = diameter_max
        self.diameter_range = diameter_max - diameter_min

    def __call__(self, alpha, beta_fraction):
        "Diameter = 2 * alpha * beta"
        beta_max = self.diameter_max / (2.0 * alpha)
        beta_min = self.diameter_min / (2.0 * alpha)
        beta_range = beta_max - beta_min
        beta = beta_min + beta_fraction * beta_range
        return beta


class ReturnFixedValue:
    "Parameter fixing class for parameter links."

    def __init__(self, value):
        self.value = value

    def __call__(self):
        return self.value
